package xpref;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.lang.instrument.IllegalClassFormatException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.net.URL;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.swing.JPanel;
import javax.swing.UnsupportedLookAndFeelException;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.xpath.XPathExpressionException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import xpref.formatter.IFormatter;
import xpref.formatter.PlainTextFormatter;
import xpref.util.CommandLineOptionException;
import xpref.util.SimpleXPath;
import xpref.util.Tupel;
import xpref.util.XMLUtils;
import xpref.util.XPrefException;

/**
 * Preferences class. This class manages all saved preferences and can display
 * the GUI to change these preferences.
 * 
 * @author Segaja
 */
public class XPref {
	/**
	 * Private singleton instance of this class.
	 */
	private static XPref instance = null;

	/**
	 * List with all registered components.
	 */
	private LinkedList<IXPref> componentsList = new LinkedList<IXPref>();

	/**
	 * DOMDocument that holds the menu XML.
	 */
	private Document menuXML = null;

	/**
	 * DOMDocument that holds the save XML.
	 */
	private Document saveXML = null;

	/**
	 * The list, that holds the default preferences.
	 */
	protected LinkedHashMap<String, LinkedList<Object>> defaultValuesList = new LinkedHashMap<String, LinkedList<Object>>();

	/**
	 * The list, that holds the preferences.
	 */
	protected LinkedHashMap<String, LinkedList<Object>> valuesList = new LinkedHashMap<String, LinkedList<Object>>();

	/**
	 * Holds the known sequences as defined in the menu XML.
	 */
	private LinkedHashMap<String, Node> sequencesList = new LinkedHashMap<String, Node>();

	/**
	 * Holds the possible id's for the command line and their notes.
	 */
	private LinkedHashMap<String, Node> optionsList = new LinkedHashMap<String, Node>();

	/**
	 * Name of the XML to save the changes to.
	 */
	private String saveXMLFileName = "preferencesSaved.xml";

	/**
	 * The list, that holds the default values per page.
	 */
	private Hashtable<String, Hashtable<String, LinkedList<String>>> defaultValueListPerPage = new Hashtable<String, Hashtable<String, LinkedList<String>>>();

	/**
	 * Instance of SimpleXPath.
	 */
	private SimpleXPath simpleXPath = SimpleXPath.getInstance();

	/**
	 * List of parsed data types. Data types in this list are valid.
	 */
	protected HashMap<String, Class<?>> parsedDataTypes = new HashMap<String, Class<?>>();

	/**
	 * Formatter for help output
	 */
	private IFormatter formatter = new PlainTextFormatter();

	/**
	 * ========== public methods to get an instance ==========
	 */

	/**
	 * Method to get an instance of the XPref class with the default path for
	 * the XML files ("/").
	 * 
	 * @return a XPref instance
	 * 
	 * @throws IOException
	 * @throws ParserConfigurationException
	 * @throws SAXException
	 * @throws XPathExpressionException
	 * @throws IllegalClassFormatException
	 * @throws NoSuchMethodException
	 * @throws NoSuchFieldException
	 * @throws ClassNotFoundException
	 * @throws SecurityException
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 * @throws IllegalArgumentException
	 */
	public static synchronized XPref getInstance() throws SAXException, ParserConfigurationException, IOException, XPathExpressionException, SecurityException, ClassNotFoundException, NoSuchFieldException, NoSuchMethodException, IllegalClassFormatException, IllegalArgumentException, IllegalAccessException, InvocationTargetException {
		return XPref.getInstance("data" + File.separator);
	}

	/**
	 * Method to get an instance of the XPref class with the specified path for
	 * the XML files.
	 * 
	 * @param xmlPath
	 *            path to the folder containing the menu and save XML file
	 * 
	 * @return a XPref instance
	 * 
	 * @throws IOException
	 * @throws ParserConfigurationException
	 * @throws SAXException
	 * @throws XPathExpressionException
	 * @throws IllegalClassFormatException
	 * @throws NoSuchMethodException
	 * @throws NoSuchFieldException
	 * @throws ClassNotFoundException
	 * @throws SecurityException
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 * @throws IllegalArgumentException
	 */
	public static synchronized XPref getInstance(String xmlPath) throws SAXException, ParserConfigurationException, IOException, XPathExpressionException, SecurityException, ClassNotFoundException, NoSuchFieldException, NoSuchMethodException, IllegalClassFormatException, IllegalArgumentException, IllegalAccessException, InvocationTargetException {
		if(XPref.instance == null) {
			XPref.instance = new XPref(xmlPath);
		}

		return XPref.instance;
	}

	/**
	 * Method to get an instance of the XPref class with the stream as the
	 * source for the XML data.
	 * 
	 * @param ressource
	 *            stream to the menu XML file
	 * 
	 * @return a XPref instance
	 * 
	 * @throws IOException
	 * @throws ParserConfigurationException
	 * @throws SAXException
	 * @throws IllegalClassFormatException
	 * @throws NoSuchMethodException
	 * @throws NoSuchFieldException
	 * @throws ClassNotFoundException
	 * @throws SecurityException
	 * @throws XPathExpressionException
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 * @throws IllegalArgumentException
	 */
	public static synchronized XPref getInstance(URL ressource) throws IOException, SAXException, ParserConfigurationException, XPathExpressionException, SecurityException, ClassNotFoundException, NoSuchFieldException, NoSuchMethodException, IllegalClassFormatException, IllegalArgumentException, IllegalAccessException, InvocationTargetException {
		if(XPref.instance == null) {
			XPref.instance = new XPref(ressource.openStream());
		}

		return XPref.instance;
	}


	/**
	 * ========== private constructors ==========
	 */

	/**
	 * Internal constructor to create an instance of the XPref class with the
	 * specified path as source of the XML files.
	 * 
	 * @param xmlPath
	 *            path to the folder containing the menu and save XML file
	 *
	 * @throws IOException
	 * @throws ParserConfigurationException
	 * @throws SAXException
	 * @throws XPathExpressionException
	 * @throws IllegalClassFormatException
	 * @throws NoSuchMethodException
	 * @throws NoSuchFieldException
	 * @throws ClassNotFoundException
	 * @throws SecurityException
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 * @throws IllegalArgumentException
	 */
	private XPref(String xmlPath) throws SAXException, ParserConfigurationException, IOException, XPathExpressionException, SecurityException, ClassNotFoundException, NoSuchFieldException, NoSuchMethodException, IllegalClassFormatException, IllegalArgumentException, IllegalAccessException, InvocationTargetException {
		this.menuXML = XMLUtils.readMenuXML(xmlPath + "preferencesMenu.xml");

		this.saveXMLFileName = xmlPath + this.saveXMLFileName;

		this.processSequences();
		this.validateDataTypes();
		this.processDefaultValues();
		this.processSaveXML(this.saveXMLFileName);
	}

	/**
	 * Internal constructor to create an instance of the XPref class with the
	 * given stream as source for the XML files.
	 * 
	 * @param xml
	 *            stream to the menu XML file
	 * 
	 * @throws IOException
	 * @throws ParserConfigurationException
	 * @throws SAXException
	 * @throws IllegalClassFormatException
	 * @throws NoSuchMethodException
	 * @throws NoSuchFieldException
	 * @throws ClassNotFoundException
	 * @throws SecurityException
	 * @throws XPathExpressionException
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 * @throws IllegalArgumentException
	 */
	private XPref(InputStream xml) throws SAXException, ParserConfigurationException, IOException, XPathExpressionException, SecurityException, ClassNotFoundException, NoSuchFieldException, NoSuchMethodException, IllegalClassFormatException, IllegalArgumentException, IllegalAccessException, InvocationTargetException {
		this.menuXML = XMLUtils.readMenuXML(xml);

		this.processSequences();
		this.validateDataTypes();
		this.processDefaultValues();
	}


	/**
	 * ========== public methods to get the GUI ==========
	 */

	/**
	 * Method to display the XPref GUI. Any changes that are made to the options
	 * will be saved to the save XML file, when the changes are applied.
	 * 
	 * @throws UnsupportedLookAndFeelException
	 * @throws IllegalAccessException
	 * @throws InstantiationException
	 * @throws ClassNotFoundException
	 * @throws XPathExpressionException
	 */
	public void showDialog() throws XPathExpressionException, ClassNotFoundException, InstantiationException, IllegalAccessException, UnsupportedLookAndFeelException {
		this.showDialog(true);
	}

	/**
	 * Method to display the XPref GUI. Any changes that are made to the options
	 * will only be saved to the save XML file, if the given boolean is set
	 * true.
	 * 
	 * @param accessToFileSystem
	 *            true, if changes should be saved to XML, when applied; false
	 *            otherwise
	 * 
	 * @throws UnsupportedLookAndFeelException
	 * @throws IllegalAccessException
	 * @throws InstantiationException
	 * @throws ClassNotFoundException
	 * @throws XPathExpressionException
	 */
	public void showDialog(boolean accessToFileSystem) throws XPathExpressionException, ClassNotFoundException, InstantiationException, IllegalAccessException, UnsupportedLookAndFeelException {
		this.showDialog(accessToFileSystem, new LinkedList<String>());
	}

	/**
	 * Method to display the XPref GUI with only the pages that contain the
	 * given id's or an page that has the given id's. Any changes that are made
	 * to the options will be saved to the save XML file, when the changes are
	 * applied.
	 * 
	 * @param idList
	 *            list of id's to be shown in the GUI
	 * 
	 * @throws UnsupportedLookAndFeelException
	 * @throws IllegalAccessException
	 * @throws InstantiationException
	 * @throws ClassNotFoundException
	 * @throws XPathExpressionException
	 */
	public void showDialog(LinkedList<String> idList) throws XPathExpressionException, ClassNotFoundException, InstantiationException, IllegalAccessException, UnsupportedLookAndFeelException {
		this.showDialog(true, idList);
	}

	/**
	 * Method to display the XPref GUI with only the pages that contain the
	 * given id's or an page that has the given id's. Any changes that are made
	 * to the options will only be saved to the save XML file, if the given
	 * boolean is set true.
	 * 
	 * @param idList
	 *            list of id's to be shown in the GUI
	 * @param accessToFileSystem
	 *            true, if changes should be saved to XML, when applied; false
	 *            otherwise
	 * 
	 * @throws UnsupportedLookAndFeelException
	 * @throws IllegalAccessException
	 * @throws InstantiationException
	 * @throws ClassNotFoundException
	 * @throws XPathExpressionException
	 */
	public void showDialog(boolean accessToFileSystem, LinkedList<String> idList) throws XPathExpressionException, ClassNotFoundException, InstantiationException, IllegalAccessException, UnsupportedLookAndFeelException {
		new XPrefGUI(this, this.componentsList, accessToFileSystem, idList, this.saveXML, this.sequencesList, this.defaultValueListPerPage, this.saveXMLFileName);
	}


	/**
	 * ========== public method to register a component ==========
	 */

	/**
	 * Method to register a component. The component to register must implement
	 * the interface IXPref. The method specified by that interface will be
	 * called on changes to the saved values.
	 * 
	 * @param component
	 *            component to register
	 */
	public synchronized void registerComponent(IXPref component) {
		this.componentsList.add(component);
	}


	/**
	 * ========== public methods to get and set the value of a given id ==========
	 */

	/**
	 * Public method which prints the internal state of the preferences object
	 * i.e. the current set of preferences.
	 * 
	 * @return a string with the current internal status
	 */
	public String toString() {
		return this.valuesList.toString();
	}

	/**
	 * Method to get the value for the given id.
	 * 
	 * @param id
	 *            the id to get the value to
	 * 
	 * @return the list of values as an instance of the proper data type
	 */
	public LinkedList<Object> getValues(String id) {
		return this.valuesList.get(id);
	}

	/**
	 * Method to get the value for the given id with the possibility to get the
	 * defaultValue.
	 * 
	 * @param id
	 *            the id to get the value to
	 * @param getDefault
	 *            boolean whether to get the default value or not
	 * 
	 * @return the list of values as an instance of the proper data type
	 */
	protected LinkedList<Object> getValues(String id, boolean getDefault) {
		if(getDefault) {
			return this.defaultValuesList.get(id);
		}
		else {
			return this.getValues(id);
		}
	}

	/**
	 * Method to set the values for the given id.
	 *
	 * @param id
	 *            the id to set the value for
	 * @param valueList
	 *            list of values to set
	 */
	public void setValues(String id, LinkedList<Object> valueList) {
		this.valuesList.put(id, valueList);
	}


	/**
	 * ========== public methods to load from and save to a file ==========
	 */

	/**
	 * Method to load settings from a specified saveXML.
	 * 
	 * @param fileName
	 *            path to the saveXML file
	 * 
	 * @throws IOException
	 * @throws ParserConfigurationException
	 * @throws SAXException
	 * @throws XPathExpressionException
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 * @throws NoSuchMethodException
	 * @throws ClassNotFoundException
	 * @throws IllegalArgumentException
	 * @throws SecurityException
	 */
	public void loadFromFile(String fileName) throws XPathExpressionException, SAXException, ParserConfigurationException, IOException, SecurityException, IllegalArgumentException, ClassNotFoundException, NoSuchMethodException, IllegalAccessException, InvocationTargetException {
		this.valuesList.clear();
		this.defaultValuesList.clear();
		this.processDefaultValues();
		this.processSaveXML(fileName);
	}

	/**
	 * Method to save the current settings to a specified saveXML.
	 * 
	 * @param fileName
	 *            path to the saveXML file
	 * 
	 * @throws ParserConfigurationException
	 * @throws XPathExpressionException
	 * @throws ClassNotFoundException
	 *             if the data type is not valid
	 * @throws NoSuchMethodException
	 *             if the data type is not valid
	 * @throws SecurityException
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 * @throws IllegalArgumentException
	 * @throws TransformerException
	 * @throws IOException
	 * @throws XPrefException
	 */
	@SuppressWarnings("unchecked")
	public void saveToFile(String fileName) throws ParserConfigurationException, XPathExpressionException, ClassNotFoundException, SecurityException, NoSuchMethodException, IllegalArgumentException, IllegalAccessException, InvocationTargetException, IOException, TransformerException, XPrefException {
		System.out.println("Saving to file...");

		if(fileName == null || fileName.isEmpty()) {
			throw new XPrefException("Invalid file name '" + fileName + "'");
		}

		// creating XML file...
		DocumentBuilderFactory documentBuilderFactory = DocumentBuilderFactory.newInstance();
		DocumentBuilder documentBuilder = documentBuilderFactory.newDocumentBuilder();

		Document saveXML = documentBuilder.newDocument();
		saveXML.appendChild(saveXML.createElement("savedPreferences"));


		LinkedHashMap<String, LinkedList<Object>> saveList = (LinkedHashMap<String, LinkedList<Object>>) this.valuesList.clone();
		HashMap<String, Tupel<Class<?>, Method>> datatypeMethods = new HashMap<String, Tupel<Class<?>, Method>>();

		// --- filter default values from save list - begin ---
		for(String pageName : this.defaultValueListPerPage.keySet()) {
			Hashtable<String, LinkedList<String>> defaultList = this.defaultValueListPerPage.get(pageName);

			for(String id : defaultList.keySet()) {
				if(this.valuesList.containsKey(id)) {
					// --- determine data type - begin ---
					String datatype = "";

					if(!id.contains(".")) { // no sequence...
						datatype = this.simpleXPath.getString("/preferences//page//option[@id='" + id + "']/@datatype", this.menuXML);
					}
					else { // sequence
						String[] idParts = id.split("\\.");

						String sequenceName = this.simpleXPath.getString("/preferences//page//option[@id='" + idParts[0] + "']/@datatype", this.menuXML);
						datatype = this.simpleXPath.getString("/preferences/sequence[@name='" + sequenceName + "']/option[@id='" + idParts[1] + "']/@datatype", this.menuXML);
					}
					// --- determine data type - end ---

					// get serialization method...
					Class<?> datatypeClass = Class.forName(datatype);
					Method serializeMethod = datatypeClass.getMethod("serializeElement", Object.class);
					datatypeMethods.put(id, new Tupel<Class<?>, Method>(datatypeClass, serializeMethod));

					boolean changed = false;

					// compare values...
					for(Object savedValueObject : this.valuesList.get(id)) {
						String savedValue = (String) serializeMethod.invoke(datatypeClass, savedValueObject);

						if(!defaultList.get(id).contains(savedValue)) {
							changed = true;

							break;
						}
					}

					if(!changed) {
						saveList.remove(id);
					}

				}
			}
		}
		// --- filter default values from save list - end ---

		// --- save the remaining values - begin ---
		// get rootNode of saveXML...
		Node rootNode = saveXML.getDocumentElement();

		// walk through saveList and create Elements for the children in
		// saveXML...
		for(String id : saveList.keySet()) {
			// get class and serialization method for current id...
			Tupel<Class<?>, Method> idStuffTupel = datatypeMethods.get(id);

			// get value objects of current id
			LinkedList<Object> values = saveList.get(id);

			StringBuffer value = new StringBuffer();

			if(values.size() == 1) {
				value.append(idStuffTupel.second.invoke(idStuffTupel.first, values.get(0)));
			}
			else {
				value.append("{{");

				for(int i = 0; i < values.size(); i += 1) {
					value.append(idStuffTupel.second.invoke(idStuffTupel.first, values.get(i)));

					if(i < values.size()-1) {
						value.append(" $$ ");
					}
				}

				value.append("}}");
			}

			Element optionNode = saveXML.createElement("item");
			optionNode.setAttribute("id", id);
			optionNode.appendChild(saveXML.createTextNode(value.toString()));

			rootNode.appendChild(optionNode);
		}

		XMLUtils.writeXMLFile(fileName, saveXML);
		// --- save the remaining values - end ---
	}

	/**
	 * Public method to parse command line options.
	 * 
	 * @param args
	 *            the command line arguments
	 * 
	 * @return the output of the parsing process.
	 * 
	 * @throws CommandLineOptionException
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 * @throws NoSuchMethodException
	 *             if the data type is not valid
	 * @throws IllegalArgumentException
	 * @throws SecurityException
	 * @throws ClassNotFoundException
	 * @throws XPathExpressionException
	 */
	public String parseCommandLineOptions(String[] args) throws CommandLineOptionException, SecurityException, IllegalArgumentException, NoSuchMethodException, IllegalAccessException, InvocationTargetException, XPathExpressionException, ClassNotFoundException {
		return this.parseCommandLineOptions(args, this.formatter);
	}

	/**
	 * Method to parse the given command line options.
	 * 
	 * @param args
	 *            The string array with the command line options
	 * 
	 * @throws CommandLineOptionException
	 *             if an error occurred
	 * 
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 * @throws NoSuchMethodException
	 *             if the data type is not valid
	 * @throws IllegalArgumentException
	 * @throws SecurityException
	 * @throws ClassNotFoundException
	 * @throws XPathExpressionException
	 */
	public String parseCommandLineOptions(String[] args, IFormatter formatter) throws CommandLineOptionException, SecurityException, IllegalArgumentException, NoSuchMethodException, IllegalAccessException, InvocationTargetException, XPathExpressionException, ClassNotFoundException {
		if(args == null || args.length == 0) {
			return ""; // nothing to do
		}

		formatter.reset();

		if(args.length == 1 && args[0].equals("--helpall")) {
			this.printHelpAll(formatter);

			return formatter.getContent();
		}

		if(args.length == 2 && args[0].equals("--help")) {
			this.printHelp(args[1], formatter);

			return formatter.getContent();
		}

		if(args.length % 2 != 0) {
			throw new CommandLineOptionException("The argument array has an odd number of elements!");
		}

		String output = "";

		for(int i = 0; i < args.length; i += 2) {
			String id = args[i].substring(2);
			String value = args[i+1];

			if(!this.optionsList.containsKey(id)) {
				throw new CommandLineOptionException("Invalid option id. See --helpall for a list of all possible options.");
			}

			Node optionNode = this.optionsList.get(id);
			String visible = XMLUtils.getAttributeValue(optionNode, "visible");

			System.out.println("visible: " + visible);

			if(visible != null && !visible.equals("") && !visible.equalsIgnoreCase("commandline")) {
				throw new CommandLineOptionException("Invalid option id. See --helpall for a list of all possible options.");
			}

			this.parseSavedValue(id, value, false);
		}

		return output;
	}

	/**
	 * Method to set a formatter for the --helpall command line option.
	 * The default formatter is xpref.formatter.PlainTextFormatter
	 *
	 * @param formatter
	 *            the new formatter
	 */
	public void setFormatter(IFormatter formatter) {
		this.formatter = formatter;
	}

	/**
	 * Method to get the menu XML.
	 * This is needed for the GUI and some data types.
	 * 
	 * @return the DOMDocument of the menu XML.
	 */
	public Document getMenuXML() {
		return this.menuXML;
	}

	/**
	 * Method to get the save XML.
	 * This is needed for the GUI.
	 * 
	 * @return the DOMDocument of the save XML.
	 */
	public Document getSaveXML() {
		return this.saveXML;
	}

	/**
	 * Method to get the separator of sequence elements.
	 * This is needed in some data types. By default this separator is " | ".
	 * 
	 * @return the separator
	 */
	public String getSequenceSeparator() {
		return " | ";
	}

	/**
	 * ========== private / protected methods ==========
	 */

	/**
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 * @throws NoSuchMethodException
	 *             if the data type is not valid
	 * @throws IllegalArgumentException
	 * @throws SecurityException
	 * @throws ClassNotFoundException
	 * @throws XPathExpressionException
	 */
	private void printHelp(String id, IFormatter formatter) throws CommandLineOptionException, SecurityException, IllegalArgumentException, NoSuchMethodException, IllegalAccessException, InvocationTargetException, XPathExpressionException, ClassNotFoundException {
		formatter.addText("\nHelp for id '" + id + "':\n");

		if(this.optionsList.containsKey(id)) {
			Node optionNode = this.optionsList.get(id);
			String nodeVisible = XMLUtils.getAttributeValue(optionNode, "visible");

			if(nodeVisible == null || nodeVisible.equals("") || nodeVisible.equalsIgnoreCase("commandline")) {
				this.printHelpDetails(id, optionNode, formatter);

				return;
			}
		}

		throw new CommandLineOptionException("No help available for this option.");
	}

	/**
	 * 
	 * @param formatter
	 * 
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 * @throws NoSuchMethodException
	 *             if the data type is not valid
	 * @throws IllegalArgumentException
	 * @throws SecurityException
	 * @throws ClassNotFoundException
	 * @throws XPathExpressionException
	 */
	private void printHelpAll(IFormatter formatter) throws SecurityException, IllegalArgumentException, NoSuchMethodException, IllegalAccessException, InvocationTargetException, XPathExpressionException, ClassNotFoundException {
		formatter.addText("\nDetails of all possible options:\n\n");

		for(String id : this.optionsList.keySet()) {
			Node optionNode = this.optionsList.get(id);
			String nodeVisible = XMLUtils.getAttributeValue(optionNode, "visible");

			if(nodeVisible != null && !nodeVisible.equals("") && !nodeVisible.equalsIgnoreCase("commandline")) {
				continue;
			}

			this.printHelpDetails(id, optionNode, formatter);
		}
	}

	/**
	 * Internal method to print the help details for the given id.
	 * 
	 * @param nodeId
	 *            the id to print the help for
	 * @param node
	 *            node of this id
	 * 
	 * @throws NoSuchMethodException
	 *             if the data type is not valid
	 * @throws SecurityException
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 * @throws IllegalArgumentException
	 * @throws ClassNotFoundException if the data type is not valid
	 * @throws XPathExpressionException
	 */
	private void printHelpDetails(String nodeId, Node optionNode, IFormatter formatter) throws SecurityException, NoSuchMethodException, IllegalArgumentException, IllegalAccessException, InvocationTargetException, XPathExpressionException, ClassNotFoundException {
		String nodeName = XMLUtils.getAttributeValue(optionNode, "name");
		String nodeDatatype = XMLUtils.getAttributeValue(optionNode, "datatype");
		String nodeHelp = XMLUtils.getAttributeValue(optionNode, "help");

		String helpText = (nodeHelp == null) ? "No help available for this option" : nodeHelp;

		StringBuffer usage = new StringBuffer();
		LinkedList<Tupel<Class<?>,String>> usageList = new LinkedList<Tupel<Class<?>, String>>();
		StringBuffer defaultValue = new StringBuffer();

		Class<?> dataTypeClass = null;
		String description = "";

		if(this.sequencesList.containsKey(nodeDatatype)) { // handle possible values for sequence
			optionNode = this.sequencesList.get(nodeDatatype);

			NodeList sequenceNodes = ((Element) optionNode).getElementsByTagName("option");

			usage.append("\"");
			defaultValue.append("\"");

			for(int i = 0; i < sequenceNodes.getLength(); i += 1) {
				usage.append(XMLUtils.getAttributeValue(sequenceNodes.item(i), "datatype"));

				if(i < sequenceNodes.getLength()-1) {
					usage.append(" | ");
				}
			}

			usage.append("\"");

			for(int i = 0; i < sequenceNodes.getLength(); i += 1) {
				String sequenceNodeDataType = XMLUtils.getAttributeValue(sequenceNodes.item(i), "datatype");

				dataTypeClass = this.parsedDataTypes.get(sequenceNodeDataType);

				Method mShowUsage = dataTypeClass.getMethod("usage", Node.class);

				String sequenceNodeUsage = (String) mShowUsage.invoke(dataTypeClass, sequenceNodes.item(i));

				usageList.add(new Tupel<Class<?>, String>(dataTypeClass, sequenceNodeUsage));


				String sequenceNodeId = nodeId + "." + XMLUtils.getAttributeValue(sequenceNodes.item(i), "id");

				Method mSerializeElement = dataTypeClass.getMethod("serializeElement", Object.class);

				defaultValue.append((String) mSerializeElement.invoke(dataTypeClass, this.getValues(sequenceNodeId).get(0)));

				if(i < sequenceNodes.getLength()-1) {
					defaultValue.append(" | ");
				}
			}

			defaultValue.append("\"");
		}
		else { // handle normal possible values
			dataTypeClass = this.parsedDataTypes.get(nodeDatatype);

			Method mShowUsage = dataTypeClass.getMethod("usage", Node.class);

			usage.append((String) mShowUsage.invoke(dataTypeClass, optionNode));

			Method mSerializeElement = dataTypeClass.getMethod("serializeElement", Object.class);

			defaultValue.append(mSerializeElement.invoke(dataTypeClass, this.getValues(nodeId).get(0)));


			Method mDescription = dataTypeClass.getMethod("description");

			description = (String) mDescription.invoke(dataTypeClass);
		}

		formatter.newBlock(nodeId, nodeName, nodeDatatype, defaultValue.toString(), helpText, usage.toString(), description, usageList);
	}

	/**
	 * Internal method to validate the data types that are used.
	 * 
	 * @throws XPathExpressionException
	 * @throws ClassNotFoundException
	 *             if the data type does not exist
	 * @throws NoSuchFieldException
	 * @throws SecurityException
	 * @throws NoSuchMethodException
	 * @throws IllegalClassFormatException
	 */
	private void validateDataTypes() throws XPathExpressionException, ClassNotFoundException, SecurityException, NoSuchFieldException, NoSuchMethodException, IllegalClassFormatException {
		System.out.println("Validating data types...");

		// get all option nodes which have a data type attribute...
		NodeList optionNodes = this.simpleXPath.getNodeSet("//option[@datatype]", this.menuXML);

		// walk through options which have a data type attribute...
		for(int i = 0; i < optionNodes.getLength(); i += 1) {

			Node optionNode = optionNodes.item(i);

			String datatype = XMLUtils.getAttributeValue(optionNode, "datatype");

			// skip sequences...
			if(this.sequencesList.containsKey(datatype)) {
				continue;
			}

			// parse each data type only ones...
			if(this.parsedDataTypes.containsKey(datatype)) {
				continue;
			}

			// get class of data type...
			Class<?> dataTypeClass = Class.forName(datatype);

			Method method = null;

			dataTypeClass.getDeclaredField("defaultValue");

			// --- check for public methods - begin ---
			method = dataTypeClass.getMethod("serializeElement", Object.class);
			this.validateModifier(method);
			this.validateReturnType(method, String.class);

			method = dataTypeClass.getMethod("parseElement", Node.class, String.class);
			this.validateModifier(method);

			method = dataTypeClass.getMethod("getVisualRepresentation", String.class, Object.class, Node.class, XPrefGUI.class, String.class, Integer.class);
			this.validateModifier(method);
			this.validateReturnType(method, JPanel.class);

			method = dataTypeClass.getMethod("getValues", String.class);
			this.validateModifier(method);
			this.validateReturnType(method, LinkedList.class);

			method = dataTypeClass.getMethod("getFirstValue", String.class);
			this.validateModifier(method);

			method = dataTypeClass.getMethod("description");
			this.validateModifier(method);
			this.validateReturnType(method, String.class);

			method = dataTypeClass.getMethod("usage", Node.class);
			this.validateModifier(method);
			this.validateReturnType(method, String.class);

			method = dataTypeClass.getMethod("getName", Node.class);
			this.validateModifier(method);
			this.validateReturnType(method, String.class);
			// --- check for public methods - end ---

			this.parsedDataTypes.put(datatype, dataTypeClass);
		}
	}

	/**
	 * Internal method to validate the modifiers of a data type method.
	 * 
	 * @param method
	 *            Method to validate
	 * 
	 * @throws IllegalClassFormatException
	 *             If the modifiers are not valid.
	 */
	private void validateModifier(Method method) throws IllegalClassFormatException {
		int modifier = method.getModifiers();

		if(!Modifier.isStatic(modifier)) {
			throw new IllegalClassFormatException("Method '" + method.getName() + "' of datatype '" + method.getClass().getName() + "' has to be static!");
		}
	}

	/**
	 * Internal method to validate the return type of a data type method.
	 * 
	 * @param method
	 *            Method to validate
	 * @param returnTypeClass
	 *            Return type the method should have
	 * 
	 * @throws IllegalClassFormatException
	 *             If the return type is not valid.
	 */
	private void validateReturnType(Method method, Class<?> returnTypeClass) throws IllegalClassFormatException {
		if(!method.getReturnType().equals(returnTypeClass)) {
			throw new IllegalClassFormatException("Method '" + method.getName() + "' of datatype '" + method.getClass().getName() + "' has to return an instance of '" + returnTypeClass.getName() + "'!");
		}
	}

	/**
	 * Load items from menuXML into the valuesList.
	 * 
	 * @throws XPathExpressionException
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 * @throws NoSuchMethodException
	 * @throws ClassNotFoundException
	 * @throws IllegalArgumentException
	 * @throws SecurityException
	 */
	protected void processDefaultValues() throws XPathExpressionException, SecurityException, IllegalArgumentException, ClassNotFoundException, NoSuchMethodException, IllegalAccessException, InvocationTargetException {
		System.out.println("Parsing default values...");

		// get list of all default values...
		NodeList defaultItems = this.simpleXPath.getNodeSet("/preferences/page//option", this.menuXML);

		// walk through list of default values...
		for(int i = 0; i < defaultItems.getLength(); i += 1) {
			Node item = defaultItems.item(i); // get current value

			String id = XMLUtils.getAttributeValue(item, "id");
			String value = XMLUtils.getAttributeValue(item, "value");
			String datatype = XMLUtils.getAttributeValue(item, "datatype");

			this.optionsList.put(id, item);

			this.parseDefaultValue(item, id, datatype, value, false);
		}
	}

	/**
	 * Internal method to parse one value.
	 * 
	 * @param optionNode
	 *            node of this option
	 * @param id
	 *            id of the option this value belongs to
	 * @param datatype
	 *            data type of the value
	 * @param value
	 *            the value string to parse
	 * 
	 * @throws ClassNotFoundException
	 *             if the data type does not exist
	 * @throws NoSuchMethodException
	 *             if the data type is not valid
	 * @throws SecurityException
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 * @throws IllegalArgumentException
	 * @throws XPathExpressionException
	 */
	private void parseDefaultValue(Node optionNode, String id, String datatype, String value, boolean isList) throws ClassNotFoundException, SecurityException, NoSuchMethodException, IllegalArgumentException, IllegalAccessException, InvocationTargetException, XPathExpressionException {
		Pattern pattern = Pattern.compile("\\{\\{(.*)\\}\\}");
		Matcher matcher = pattern.matcher(value);

		if(matcher.find() == true) { // option has a list of values...
			isList = true;

			// get list elements...
			String[] listElements = matcher.group(1).split(" \\$\\$ ");


			// get minimum amount...
			int minimum = 1;

			String minString = XMLUtils.getAttributeValue(optionNode, "minimum");

			if(minString != null) { // minimum provided...
				minimum = Integer.parseInt(minString);
			}
			else { // no minimum provided...
				System.err.println("WARNING: The option with the id '" + id + "' has no 'minimum' attribute specified for its list! A minimum of 1 will be assumed!");
			}


			// check lower bound...
			if(listElements.length < minimum) { // the lower limit is higher then the amount of elements we have...
				int missing = minimum - listElements.length; // amount of missing elements

				System.err.println("WARNING: The option with the id '" + id + " has a higher 'minimum' attribute then elements for the list! '" + missing + "' (minimum - amount of elements in the list) elements will be added with the data type default value to the list!");

				// create new list which is big enough...
				String[] newListElements = new String[listElements.length + missing];

				// copy elements from the old list into the new...
				for(int i = 0; i < listElements.length; i += 1) {
					newListElements[i] = listElements[i];
				}

				// add amount of missing elements...
				for(int i = 0; i < missing; i += 1) {
					Class<?> c = Class.forName(datatype); // get class

					Object obj = null;

					Method m = c.getMethod("serializeElement", Object.class); // get method
					String valueString = (String) m.invoke(c, obj); // invoke method

					newListElements[listElements.length] = valueString;
				}

				// write new list back into variable of the old list...
				listElements = newListElements;
			}


			// get maximum amount...
			int maximum = -1;

			String maxString = XMLUtils.getAttributeValue(optionNode, "maximum");

			if(maxString != null) { // maximum provided...
				maximum = Integer.parseInt(maxString);
			}
			else { // maximum not provided...
				System.err.println("WARNING: The option with the id '" + id + "' has no 'maximum' attribute specified for its list! The whole list will be used!");
			}


			// check upper bound...
			if(maximum == -1) { // no upper limit so the whole list will be parsed...
				maximum = listElements.length;
			}
			else if(maximum < listElements.length) { // there are more list elements then the upper limit allows...
				System.err.println("WARNING: The option with the id '" + id + "' has more elements in its list then the 'maximum' of the option allows! Only the first " + maximum + " (value of maximum attribut) elements will be considered!");
			}


			for(String listElement : listElements) {
				this.parseDefaultValue(optionNode, id, datatype, listElement, isList);
			}
		}
		else if(this.sequencesList.containsKey(datatype)) { // option is a sequence...
			Node sequenceNode = this.sequencesList.get(datatype); // get sequence node

			if(value == null || value.equals("")) { // no default values for this sequence given...
				System.err.println("WARNING: No default values for the sequence '" + id + "' specified! Reverting back to default values of the sequence definition (" + datatype + ")!");

				this.retrieveSequenceDefaultValues(id, sequenceNode, optionNode);

				return;
			}

			// test amount of default values for the sequence...
			String[] defaultValueElements = value.split(" \\| ");

			NodeList optionsInSequenceList =  this.simpleXPath.getNodeSet("./option", sequenceNode);

			if(defaultValueElements.length != optionsInSequenceList.getLength()) {
				System.err.println("WARNING: Amount of default values for sequence '" + id + "' does not match amount of elements in sequence definition! Reverting back to default values of the sequence definition (" + datatype + ")!");

				this.retrieveSequenceDefaultValues(id, sequenceNode, optionNode);

				return;
			}

			// parsing sequence default values...
			for(int k = 0; k < optionsInSequenceList.getLength(); k += 1) {
				Node sequenceOptionNode = optionsInSequenceList.item(k); // get current option node

				String soID = id + "." + XMLUtils.getAttributeValue(sequenceOptionNode, "id");
				String soDatatype = XMLUtils.getAttributeValue(sequenceOptionNode, "datatype");

				this.addDefaultValue(optionNode, soID, defaultValueElements[k]);
				this.addNodeValue(soID, soDatatype, defaultValueElements[k], isList, true);
			}
		}
		else { // option is a normal value...
			this.addDefaultValue(optionNode, id, value);
			this.addNodeValue(id, datatype, value, isList, true);
		}
	}

	private void addDefaultValue(Node optionNode, String id, String value) {
		String pageName = this.findPageName(optionNode);

		if(!this.defaultValueListPerPage.containsKey(pageName)) { // page not yet present...
			this.defaultValueListPerPage.put(pageName, new Hashtable<String, LinkedList<String>>());
		}

		Hashtable<String, LinkedList<String>> pageList = this.defaultValueListPerPage.get(pageName);

		if(!pageList.containsKey(id)) {
			pageList.put(id, new LinkedList<String>());
		}

		LinkedList<String> idList = pageList.get(id);
		idList.add(value);
	}

	private String findPageName(Node optionNode) {
		Node node = optionNode.getParentNode();

		if(node.getNodeName().equals("page")) {
			return XMLUtils.getAttributeValue(node, "name");
		}
		else {
			return this.findPageName(node);
		}
	}

	/**
	 * Internal method to get all defined sequences as list.
	 * 
	 * @throws XPathExpressionException
	 */
	private void processSequences() throws XPathExpressionException {
		NodeList sequencesItems = this.simpleXPath.getNodeSet("/preferences/sequence", this.menuXML);

		for(int i = 0; i < sequencesItems.getLength(); i += 1) {
			Node sequenceNode = sequencesItems.item(i);

			this.sequencesList.put(XMLUtils.getAttributeValue(sequenceNode, "name"), sequenceNode);
		}
	}

	/**
	 * Internal method to add a given value of a given ID with the given data
	 * type to the values list.
	 * 
	 * @param id
	 *            id of value to add
	 * @param datatype
	 *            data type of value to add
	 * @param value
	 *            value to add
	 * @param isList
	 *            true if the id has a list of values
	 * @param isDefault
	 *            boolean whether the value is a default value or not
	 * 
	 * @throws XPathExpressionException
	 * @throws ClassNotFoundException
	 *             if data type does not exist
	 * @throws NoSuchMethodException
	 *             if data type is invalid
	 * @throws SecurityException
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 * @throws IllegalArgumentException
	 */
	private void addNodeValue(String id, String datatype, String value, boolean isList, boolean isDefault) throws XPathExpressionException, ClassNotFoundException, SecurityException, NoSuchMethodException, IllegalArgumentException, IllegalAccessException, InvocationTargetException {
		Node optionNode = null;

		String[] idparts = id.split("\\.");

		if(idparts.length == 2) {
			String sequenceName = SimpleXPath.getInstance().getString("/preferences/page//option[@id='" + idparts[0] + "']/@datatype", this.menuXML);

			optionNode = SimpleXPath.getInstance().getNode("/preferences/sequence[@name='" + sequenceName + "']/option[@id='" + idparts[1] + "']", this.menuXML);
		}
		else {
			optionNode = SimpleXPath.getInstance().getNode("//option[@id='" + id + "']", this.menuXML);
		}

		if(!isDefault && XMLUtils.getAttributeValue(optionNode, "value").equals(value)) {
			isDefault = true;
		}

		Class<?> c = Class.forName(datatype); // get class

		Method m = c.getMethod("parseElement", Node.class, String.class); // get method
		Object valueObject = m.invoke(c, optionNode, value); // invoke method

		// --- add to valuesList - begin ---
		// create list for values for this id, if it does not exist yet...
		if(!this.valuesList.containsKey(id)) {
			this.valuesList.put(id, new LinkedList<Object>());
		}

		if(!isList) {
			this.valuesList.get(id).clear();
		}

		this.valuesList.get(id).add(valueObject); // store value
		// --- add to valuesList - end ---

		if(isDefault) {
			// --- add to defaultValuesList - begin ---
			// create list for values for this id, if it does not exist yet...
			if(!this.defaultValuesList.containsKey(id)) {
				this.defaultValuesList.put(id, new LinkedList<Object>());
			}

			if(!isList) {
				this.defaultValuesList.get(id).clear();
			}

			this.defaultValuesList.get(id).add(valueObject); // store value
			// --- add to defaultValuesList - end ---
		}
	}

	/**
	 * Internal method to retrieve the default values of the given sequence node
	 * an add them to the value list.
	 * 
	 * @param nodeID
	 *            id of the current node to get the values for
	 * @param sequenceNode
	 *            sequence node to get default values for
	 * 
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 * @throws NoSuchMethodException
	 * @throws ClassNotFoundException
	 * @throws IllegalArgumentException
	 * @throws SecurityException
	 * @throws XPathExpressionException
	 */
	private void retrieveSequenceDefaultValues(String nodeID, Node sequenceNode, Node optionNode) throws XPathExpressionException, SecurityException, IllegalArgumentException, ClassNotFoundException, NoSuchMethodException, IllegalAccessException, InvocationTargetException {
		NodeList sequenceNodes = sequenceNode.getChildNodes(); // get child nodes

		// walk through child nodes...
		for(int i = 0; i < sequenceNodes.getLength(); i += 1) {
			Node child = sequenceNodes.item(i); // get current child

			// only bother about "option" nodes...
			if(child.getNodeName().equals("option")) {
				String id = nodeID + "." + XMLUtils.getAttributeValue(child, "id");
				String datatype = XMLUtils.getAttributeValue(child, "datatype");
				String value = XMLUtils.getAttributeValue(child, "value");

				// add node value to values list...
				this.addNodeValue(id, datatype, value, false, false);

				if(optionNode != null) {
					this.addDefaultValue(optionNode, id, value);
				}
			}
		}
	}

	/**
	 * Internal method to parse the save XML.
	 * 
	 * @param saveXMLPath
	 *            the path to the save XML
	 * 
	 * @throws IOException
	 * @throws ParserConfigurationException
	 * @throws SAXException
	 * @throws XPathExpressionException
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 * @throws NoSuchMethodException
	 * @throws ClassNotFoundException
	 * @throws IllegalArgumentException
	 * @throws SecurityException
	 */
	private void processSaveXML(String saveXMLPath) throws SAXException, ParserConfigurationException, IOException, XPathExpressionException, SecurityException, IllegalArgumentException, ClassNotFoundException, NoSuchMethodException, IllegalAccessException, InvocationTargetException {
		this.saveXML = XMLUtils.readSaveXML(saveXMLPath);

		System.out.println("Parsing saved values...");

		// get list of all saved elements...
		NodeList savedItems = this.simpleXPath.getNodeSet("/savedPreferences/item", this.saveXML);

		// walk through list of saved elements...
		for(int i = 0; i < savedItems.getLength(); ++i) {
			Node item = savedItems.item(i); // get current element

			// get id and value for current element...
			String id = XMLUtils.getAttributeValue(item, "id");
			String value = item.getFirstChild().getNodeValue();

			this.parseSavedValue(id, value, false);
		}
	}

	/**
	 * Internal method to parse a saved value.
	 * 
	 * @param id
	 *            id of the parsed value
	 * @param value
	 *            value to parse
	 * 
	 * @throws ClassNotFoundException
	 *             if data type does not exist
	 * @throws NoSuchMethodException
	 *             if data type is invalid
	 * @throws SecurityException
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 * @throws IllegalArgumentException
	 * @throws XPathExpressionException
	 */
	protected void parseSavedValue(String id, String value, boolean isList) throws ClassNotFoundException, SecurityException, NoSuchMethodException, IllegalArgumentException, IllegalAccessException, InvocationTargetException, XPathExpressionException {
		Pattern pattern = Pattern.compile("\\{\\{(.*)\\}\\}");
		Matcher matcher = pattern.matcher(value);

		// find data type to current value...
		String datatype = this.simpleXPath.getString("/preferences//page//option[@id='" + id + "']/@datatype", this.menuXML);

		if(matcher.find() == true) { // option has a list of values...
			isList = true;

			// get list elements...
			String[] listElements = matcher.group(1).split(" \\$\\$ ");

			Node defaultNode = this.simpleXPath.getNode("//option[@id='" + id + "']", this.menuXML);

			String defaultValue = this.simpleXPath.getString("./@value", defaultNode);

			Matcher defaultMatcher = pattern.matcher(defaultValue);

			if(defaultMatcher.find() == true) { // we really have a list...
				// get minimum amount...
				int minimum = 1;

				String minString = XMLUtils.getAttributeValue(defaultNode, "minimum");

				if(minString != null) { // minimum provided...
					minimum = Integer.parseInt(minString);
				}

				// check lower bound...
				if(listElements.length < minimum) { // the lower limit is higher then the amount of elements we have...
					int missing = minimum - listElements.length; // amount of missing elements

					// create new list which is big enough...
					String[] newListElements = new String[listElements.length + missing];

					// copy elements from the old list into the new...
					for(int i = 0; i < listElements.length; i += 1) {
						newListElements[i] = listElements[i];
					}

					// get data type...
					datatype = XMLUtils.getAttributeValue(defaultNode, "datatype");

					// add amount of missing elements...
					for(int i = 0; i < missing; i += 1) {
						Class<?> c = Class.forName(datatype); // get class

						Object obj = null;

						Method m = c.getMethod("serializeElement", Object.class); // get method
						String valueString = (String) m.invoke(c, obj); // invoke method

						newListElements[listElements.length] = valueString;
					}

					// write new list back into variable of the old list...
					listElements = newListElements;
				}


				// get maximum amount...
				int maximum = -1;

				String maxString = XMLUtils.getAttributeValue(defaultNode, "maximum");

				if(maxString != null) { // maximum provided...
					maximum = Integer.parseInt(maxString);
				}

				// check upper bound...
				if(maximum == -1) { // no upper limit so the whole list will be parsed...
					maximum = listElements.length;
				}
				else if(maximum < listElements.length) { // there are more list elements then the upper limit allows...
					System.err.println("WARNING: The option with the id '" + id + "' has more elements in its list then the 'maximum' of the option allows! Only the first " + maximum + " (value of maximum attribut) elements will be considered!");
				}

				this.valuesList.remove(id); // remove the default value

				for(String listElement : listElements) {
					this.parseSavedValue(id, listElement, isList);
				}
			}
			else { // the default value is no list, so we have no list...
				System.err.println("WARNING: Saved value for id '" + id + "' is a list, but this option is no list! Only first value of the save value list is considered!");

				this.parseSavedValue(id, listElements[0], isList);
			}
		}
		else if(this.sequencesList.containsKey(datatype)) { // option is a sequence...
			Node sequenceNode = this.sequencesList.get(datatype); // get sequence node

			if(value == null) { // no default values for this sequence given...
				System.err.println("WARNING: No default values for the sequence '" + id + "' specified! Reverting back to default values of the sequence definition (" + datatype + ")!");

				this.retrieveSequenceDefaultValues(id, sequenceNode, null);

				return;
			}

			// test amount of default values for the sequence...
			String[] defaultValueElements = value.split(" \\| ");

			NodeList optionsInSequenceList =  this.simpleXPath.getNodeSet("./option", sequenceNode);

			if(defaultValueElements.length != optionsInSequenceList.getLength()) {
				System.err.println("WARNING: Amount of default values for sequence '" + id + "' does not match amount of elements in sequence definition! Reverting back to default values of the sequence definition (" + datatype + ")!");

				this.retrieveSequenceDefaultValues(id, sequenceNode, null);

				return;
			}

			// parsing sequence default values...
			for(int k = 0; k < optionsInSequenceList.getLength(); k += 1) {
				Node sequenceOptionNode = optionsInSequenceList.item(k); // get current option node

				String soID = id + "." + XMLUtils.getAttributeValue(sequenceOptionNode, "id");
				String soDatatype = XMLUtils.getAttributeValue(sequenceOptionNode, "datatype");

				this.addNodeValue(soID, soDatatype, defaultValueElements[k], isList, true);
			}
		}
		else {
			if(id.contains(".")) { // handle sequences...
				String preID = id.split("\\.")[0];
				String postID = id.split("\\.")[1];
				String sequenceName = this.simpleXPath.getString("/preferences//page//option[@id='" + preID + "']/@datatype", this.menuXML);

				datatype = this.simpleXPath.getString("/preferences/sequence[@name='" + sequenceName + "']/option[@id='" + postID + "']/@datatype", this.menuXML);
			}

			matcher = pattern.matcher(value);

			if(matcher.find() == true) { // default value is a list...
				System.err.println("WARNING: Default value for id '" + id + "' is a list, but this saved value is no list! This saved value will be interpreted as a one element list!");
			}

			this.addNodeValue(id, datatype, value, isList, false); // add the saved value
		}
	}




	//	protected void printValuesList() { // FIXME: DEBUG
	//		System.out.println();
	//
	//		for(String id : this.valuesList.keySet()) {
	//			System.out.println(id + " : " + this.valuesList.get(id).toString());
	//		}
	//
	//		System.out.println();
	//	}

	//	protected void printDefaultValuesList() { // FIXME: DEBUG
	//		System.out.println();
	//
	//		for(String id : this.defaultValuesList.keySet()) {
	//			System.out.println(id + " : " + this.defaultValuesList.get(id).toString());
	//		}
	//
	//		System.out.println();
	//	}

	//	private void printDefaultValuesPerPageList() { // FIXME: DEBUG
	//		System.out.println();
	//
	//		for(String pageName : this.defaultValueListPerPage.keySet()) {
	//			System.out.println(pageName);
	//
	//			for(String id : this.defaultValueListPerPage.get(pageName).keySet()) {
	//				System.out.println(" > " + id + " : " + this.defaultValueListPerPage.get(pageName).get(id).toString());
	//			}
	//		}
	//
	//		System.out.println();
	//	}



	/**
	 * Main method to start the XPref object and it's GUI stand alone.
	 */
	public static void main(String[] args) {
		try {
			XPref xPref = XPref.getInstance("data/");

			//			HTMLFormatter formatter = new HTMLFormatter(xPref);
			//			HTMLFormatter formatter = new HTMLFormatter("XPref Helpall", ".idlabel {color:#FF0000;}");
			//			formatter.setShortenDataTypes(true);
			//			xPref.setFormatter(formatter);
			//			System.out.println(xPref.parseCommandLineOptions(args));

			//			xPref.saveToFile("/home/segaja/xpref.xml");
			//			xPref.loadFromFile("/home/segaja/xpref.xml");
			xPref.showDialog();
		}
		catch(Exception e) {
			e.printStackTrace();
		}
	}
}