package xpref;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.FlowLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.swing.Box;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSeparator;
import javax.swing.JSplitPane;
import javax.swing.JTextField;
import javax.swing.SwingConstants;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;
import javax.swing.UnsupportedLookAndFeelException;
import javax.swing.border.LineBorder;
import javax.swing.border.TitledBorder;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.filechooser.FileNameExtensionFilter;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.xpath.XPathExpressionException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import xpref.util.History;
import xpref.util.SaveDialog;
import xpref.util.SimpleXPath;
import xpref.util.Tree;
import xpref.util.TreeNode;
import xpref.util.XMLUtils;

/**
 * GUI for XPref.
 */
public class XPrefGUI extends JFrame {
	/**
	 * Serial version UID.
	 */
	private static final long serialVersionUID = 8887298826387291184L;

	/**
	 * Instance of SimpleXPath.
	 */
	private SimpleXPath simpleXPath = SimpleXPath.getInstance();

	/**
	 * List of changed values grouped by the page they are on.
	 */
	private Hashtable<String, Hashtable<String, LinkedList<String>>> changeList = new Hashtable<String, Hashtable<String, LinkedList<String>>>();

	/**
	 * Boolean whether the GUI is the only one or not.
	 */
	private boolean standAlone;

	/**
	 * List of dependencies, grouped by the id, the depend on, the value that is a true value and the JComponent that depends on that.
	 */
	private Hashtable<String, Hashtable<String, LinkedList<JComponent>>> dependencyList = new Hashtable<String, Hashtable<String, LinkedList<JComponent>>>();

	/**
	 * Boolean whether XPref has access to the file system or not.
	 */
	private boolean accessToFileSystem;

	/**
	 * Holds the menu XML DOMDocument.
	 */
	private Document menuXML;

	/**
	 * Holds the IDs of the pages from the menu XML that should be viewed.
	 */
	private LinkedList<String> viewedPages = new LinkedList<String>();

	/**
	 * Holds the cached option panels.
	 */
	public Hashtable<String, JPanel> optionPanels = new Hashtable<String, JPanel>();

	/**
	 * JSplitPane for the window.
	 */
	private JSplitPane mainSP;

	/**
	 * Holds the known sequences as defined in the menu XML.
	 */
	private LinkedHashMap<String, Node> sequencesList = new LinkedHashMap<String, Node>();

	/**
	 * Holds the minus buttons per id.
	 */
	private HashMap<String, LinkedList<JButton>> minusButtonList = new HashMap<String, LinkedList<JButton>>();

	/**
	 * Holds the plus button per id
	 */
	private HashMap<String, JButton> plusButtonList = new HashMap<String, JButton>();

	/**
	 * List with all registered components.
	 */
	private LinkedList<IXPref> componentsList = new LinkedList<IXPref>();

	/**
	 * Name of the XML to save the changes to.
	 */
	private String saveXMLFileName = "preferencesSaved.xml";

	/**
	 * DOMDocument object that holds the save XML.
	 */
	private Document saveXML = null;

	/**
	 * The list, that holds the default values per page.
	 */
	private Hashtable<String, Hashtable<String, LinkedList<String>>> defaultValueList = new Hashtable<String, Hashtable<String, LinkedList<String>>>();

	/**
	 * Reference to the XPref object.
	 */
	private XPref xPref = null;

	/**
	 * Reference to itself.
	 */
	private XPrefGUI that = this;

	/**
	 * The Tree instance of the GUI.
	 */
	private Tree tree = null;

	/**
	 * Default root node of the tree.
	 */
	private TreeNode defaultRootNode = null;

	/**
	 * JScrollPane of the tree.
	 */
	private JScrollPane treeScrollPane = null;

	/**
	 * keyword for the tree filter.
	 */
	private String keyword = "";

	/**
	 * Instance of the history.
	 */
	private History history = null;

	/**
	 * ========== package scoped constructor ==========
	 */

	/**
	 * Package scoped constructor to display the XPrefGUI.
	 * 
	 * @param idList
	 *            List of IDs to display (if empty the all elements will be
	 *            displayed)
	 * 
	 * @throws UnsupportedLookAndFeelException
	 * @throws IllegalAccessException
	 * @throws InstantiationException
	 * @throws ClassNotFoundException
	 * @throws XPathExpressionException
	 */
	protected XPrefGUI(XPref xPref, LinkedList<IXPref> componentsList, boolean accessToFileSystem, LinkedList<String> idList, Document saveXML, LinkedHashMap<String, Node> sequencesList, Hashtable<String, Hashtable<String, LinkedList<String>>> defaultValueList, String saveXMLFileName) throws ClassNotFoundException, InstantiationException, IllegalAccessException, UnsupportedLookAndFeelException, XPathExpressionException {
		this.secure();

		this.xPref = xPref;
		this.componentsList = componentsList;
		this.standAlone = componentsList.isEmpty();
		this.accessToFileSystem = accessToFileSystem;
		this.menuXML = xPref.getMenuXML();
		this.saveXML = saveXML;
		this.sequencesList = sequencesList;
		this.defaultValueList = defaultValueList;
		this.saveXMLFileName = saveXMLFileName;

		// set system LAF if there is no component registered ...
		// (if there is one or more component registered, their LAF will be inherited by Java)
		if(this.standAlone) {
			UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
		}

		// parse the list of IDs which should be shown to determine which pages will be shown...
		this.parseIDList(idList);

		// generate panel for the tree and the search field...
		GridBagConstraints leftGBC = new GridBagConstraints();
		leftGBC.anchor = GridBagConstraints.NORTH;
		leftGBC.gridx = leftGBC.gridy = 0;
		leftGBC.gridwidth = leftGBC.gridheight = 1;
		leftGBC.weightx = 1.0;
		leftGBC.weighty = 0;
		leftGBC.fill = GridBagConstraints.HORIZONTAL;
		leftGBC.insets = new Insets(10, 10, 10, 10);

		JPanel leftPanel = new JPanel();
		leftPanel.setLayout(new GridBagLayout());


		// generate search field...
		GridBagConstraints searchGBC = new GridBagConstraints();
		searchGBC.anchor = GridBagConstraints.NORTH;
		searchGBC.gridx = searchGBC.gridy = 0;
		searchGBC.gridwidth = searchGBC.gridheight = 1;
		searchGBC.weightx = searchGBC.weighty = 1.0;
		searchGBC.fill = GridBagConstraints.BOTH;
		searchGBC.insets = new Insets(2, 2, 2, 2);

		final JTextField searchField = new JTextField("type filter text");
		searchField.setSelectionStart(0);
		searchField.setSelectionEnd(searchField.getText().length());
		searchField.addKeyListener(new KeyListener() {
			public void keyReleased(KeyEvent ke) {
				that.keyword = searchField.getText().toLowerCase();

				that.history.clear();

				if(that.keyword.equals("")) { // no filtering...
					that.tree.showNewTree(that.defaultRootNode);
				}
				else { // filtering
					that.viewedPages.clear();

					this.findMatchingPages(that.defaultRootNode);

					if(that.viewedPages.size() == 0) {
						that.tree.showNewTree(that.defaultRootNode);
					}
					else {
						try {
							TreeNode rootNode = new TreeNode("root", "", null);

							// re-generate the tree...
							that.generateBranch(rootNode, that.simpleXPath.getNodeSet("/preferences/page", that.menuXML));

							that.tree.showNewTree(rootNode);
						}
						catch(XPathExpressionException e) {
							e.printStackTrace();
						}
					}
				}
			}

			private void findMatchingPages(TreeNode node) {
				// node is not the root of the tree root and key word matches...
				if(!node.getPageID().equals("") && node.toString().toLowerCase().contains(that.keyword)) {
					that.addPageToViewedPages(node.getXMLNode());
				}

				for(int i = 0; i < node.getChildCount(); i += 1) {
					this.findMatchingPages((TreeNode) node.getChildAt(i));
				}
			}

			public void keyPressed(KeyEvent ke) {}
			public void keyTyped(KeyEvent ke) {}
		});

		JPanel searchPanel = new JPanel(new GridBagLayout());
		searchPanel.add(searchField, searchGBC);
		leftPanel.add(searchPanel, leftGBC);

		// create root node for JTree...
		leftGBC.gridy++;
		leftGBC.weighty = 1.0;
		leftGBC.anchor = GridBagConstraints.SOUTH;
		leftGBC.fill = GridBagConstraints.BOTH;
		leftGBC.insets = new Insets(0, 0, 0, 0);

		this.defaultRootNode = new TreeNode("root", "", null);
		this.tree = this.generateTree(this.defaultRootNode); // generate JTree
		this.treeScrollPane = new JScrollPane(this.tree);

		this.history = new History(this.tree);

		leftPanel.add(this.treeScrollPane, leftGBC);

		this.mainSP = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT);
		this.mainSP.setDividerLocation(this.tree.getWidth() + 4);
		this.mainSP.setOneTouchExpandable(true);
		this.mainSP.setContinuousLayout(true);
		this.mainSP.setLeftComponent(leftPanel);
		this.mainSP.setRightComponent(new JPanel());

		// set first element of JTree as selected element...
		this.tree.selectFirstChild();

		JPanel buttonPanel = new JPanel(new FlowLayout(FlowLayout.RIGHT));

		if(this.accessToFileSystem) {
			JButton importButton = new JButton("Import Settings");
			importButton.setMargin(new Insets(3, 20, 3, 20));
			importButton.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent ae) {
					JFileChooser chooser = new JFileChooser();
					chooser.setFileFilter(new FileNameExtensionFilter("XML files", "xml"));

					int ret = chooser.showDialog(that, "Load");

					if(ret == JFileChooser.APPROVE_OPTION) {
						String fileName = chooser.getSelectedFile().getAbsolutePath();

						try {
							// clear the option panel cache...
							that.optionPanels.clear();

							// clear the old changed list...
							that.changeList.clear();

							// load the new settings...
							that.xPref.loadFromFile(fileName);
							that.saveXML = that.xPref.getSaveXML();

							// --- reload the current page - begin ---
							TreeNode treeNode = that.tree.getCurrentNode();
							String pageID = treeNode.getPageID();
							String pageName = treeNode.toString();

							that.optionPanels.put(pageID, that.generateOptionPanel(pageID, pageName, false));
							that.updateRightComponent(pageID);
							// --- reload the current page - end ---

							// --- pre load the changed pages - begin ---
							NodeList itemList = that.simpleXPath.getNodeSet("/savedPreferences/item", that.saveXML);

							for(int i = 0; i < itemList.getLength(); i += 1) {
								String itemID = XMLUtils.getAttributeValue(itemList.item(i), "id");
								Node pageNode = that.simpleXPath.getNode("//page[./options//option[@id='" + itemID + "']]", that.menuXML);
								pageID = XMLUtils.getAttributeValue(pageNode, "id");
								pageName = XMLUtils.getAttributeValue(pageNode, "name");

								if(!that.optionPanels.containsKey(pageID)) {
									try {
										that.optionPanels.put(pageID, that.generateOptionPanel(pageID, pageName, false));
									}
									catch(Exception e) {
										e.printStackTrace();
									}
								}
							}
							// --- pre load the changed pages - end ---
						}
						catch(Exception e) {
							e.printStackTrace();
						}
					}
				}
			});

			JButton exportButton = new JButton("Export Settings");
			exportButton.setMargin(new Insets(3, 20, 3, 20));
			exportButton.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					SaveDialog chooser = new SaveDialog();
					chooser.setFileFilter(new FileNameExtensionFilter("XML files", "xml"));

					int ret = chooser.showDialog(that, "Save");

					if(ret == JFileChooser.APPROVE_OPTION) {
						String fileName = chooser.getSelectedFile().getAbsolutePath();

						if(!fileName.endsWith(".xml")) {
							fileName += ".xml";
						}

						try {
							that.xPref.saveToFile(fileName);
						}
						catch(Exception e1) {
							e1.printStackTrace();
						}
					}
				}
			});

			buttonPanel.add(importButton);
			buttonPanel.add(exportButton);
			buttonPanel.add(Box.createHorizontalStrut(20));
		}

		JButton okButton = new JButton("OK");
		okButton.setMargin(new Insets(3, 20, 3, 20));
		okButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				try {
					saveAllPages();
				}
				catch(Exception e1) {
					e1.printStackTrace();
				}

				exit();
			}
		});

		JButton cancelButton = new JButton("Cancel");
		cancelButton.setMargin(new Insets(3, 20, 3, 20));
		cancelButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent arg0) {
				exit();
			}
		});

		buttonPanel.add(okButton);
		buttonPanel.add(cancelButton);

		JPanel masterPanel = new JPanel(new BorderLayout());
		masterPanel.add(this.mainSP, BorderLayout.CENTER);
		masterPanel.add(buttonPanel, BorderLayout.SOUTH);

		// set some frame properties...
		this.setTitle("XPref GUI");
		this.getContentPane().add(masterPanel);
		this.setSize(900, 700);
		this.setLocationRelativeTo(null);
		this.setVisible(true);
		this.addWindowListener(new WindowAdapter() {
			public void windowClosing(WindowEvent e) {
				exit();
			}
		});
	}

	/**
	 * ========== private methods ==========
	 */

	/**
	 * Internal method to secure against wrong usage.
	 */
	private void secure() {
		boolean allow = false;

		if((new Exception()).getStackTrace()[2].getClassName().startsWith("xpref.XPref")) {
			allow = true;
		}

		allow = true;

		if(!allow) {
			throw new UnsupportedOperationException("This call is not allowed");
		}
	}

	/**
	 * Internal method to close the GUI. If there are registered components just
	 * hide the GUI. Otherwise stop the VM.
	 */
	private void exit() {
		if(!this.standAlone) {
			this.setVisible(false);
		}
		else {
			System.exit(0);
		}
	}

	/**
	 * Internal method to change the status of a Container and all its children
	 * and their children recursively.
	 * 
	 * @param component
	 *            the current container to change the status for
	 * @param newStatus
	 *            the new status
	 */
	private void changeComponentState(Container component, boolean newStatus) {
		// update status of current element...
		component.setEnabled(newStatus);

		// walk through children of current element...
		for(int i = 0; i < component.getComponentCount(); i += 1) {
			// get child of current element...
			Component child = component.getComponent(i);

			// if child is a Container...
			if(child instanceof Container) {
				// change state of the child recursively...
				this.changeComponentState((Container) child, newStatus);
			}
		}
	}

	/**
	 * Internal method to determine the pages that should be displayed depending
	 * on the given IDs. The page in which a given ID occurs will be shown. Also
	 * the path from the root page to the determined page will be shown.
	 * 
	 * @param idList
	 *            list of IDs that define the pages that should be shown.
	 * 
	 * @throws XPathExpressionException
	 */
	private void parseIDList(LinkedList<String> idList) throws XPathExpressionException {
		// add pages for this IDs to the list...
		for(String id : idList) {
			Node pageNode = this.simpleXPath.getNode("//page[./options//option[@id='" + id + "']]", this.menuXML);

			this.addPageToViewedPages(pageNode);
		}

		// no page was added... show all pages...
		if(this.viewedPages.size() == 0) {
			this.addAllPagesToView();
		}
	}

	/**
	 * Internal method to add all pages to the view.
	 * 
	 * @throws XPathExpressionException
	 */
	private void addAllPagesToView() throws XPathExpressionException {
		NodeList pageNodes = this.simpleXPath.getNodeSet("//page", this.menuXML);

		for(int i = 0; i < pageNodes.getLength(); ++i) {
			this.viewedPages.add(XMLUtils.getAttributeValue(pageNodes.item(i), "id"));
		}
	}

	/**
	 * Internal method to add the given page and the path to it to the list of
	 * displayed pages.
	 * 
	 * @param pageNode
	 *            the node of the page to add
	 */
	private void addPageToViewedPages(Node pageNode) {
		// get id of the page and add it...
		this.viewedPages.add(XMLUtils.getAttributeValue(pageNode, "id"));

		// get parent node...
		Node parentNode = pageNode.getParentNode();

		// if parent node is a page, add it...
		if(parentNode.getNodeName().equalsIgnoreCase("page")) {
			this.addPageToViewedPages(parentNode);
		}
	}

	/**
	 * Internal method to build the JTree for the pages.
	 * 
	 * @param rootNode
	 *            root node of the JTree
	 * 
	 * @return the complete JTree
	 * 
	 * @throws XPathExpressionException
	 */
	private Tree generateTree(TreeNode rootNode) throws XPathExpressionException {
		// get children of root node...
		NodeList xmlNodeList = this.simpleXPath.getNodeSet("/preferences/page", this.menuXML);

		this.generateBranch(rootNode, xmlNodeList); // generate branch from children

		Tree tree = new Tree(rootNode);
		tree.addTreeSelectionListener(new TreeSelectionListener() {
			public void valueChanged(TreeSelectionEvent tse) {
				// get selected item...
				TreeNode treeNode = that.tree.setCurrentNode(tse);
				String itemID = treeNode.getPageID();
				String itemName = treeNode.toString().replace("<html><b>", "").replace("</b></html>", "");

				// remember TreePath to this page...
				that.tree.addTreePath(itemID, treeNode.getTreePath());

				// create option panel and add it to list, if it is not in it...
				if(!that.optionPanels.containsKey(itemID)) {
					try {
						that.optionPanels.put(itemID, that.generateOptionPanel(itemID, itemName, false));
					}
					catch(Exception e) {
						e.printStackTrace();
					}
				}

				that.updateRightComponent(itemID);

				if(!treeNode.selected) {
					that.history.update(itemID);
				}
			}
		});

		return tree;
	}

	/**
	 * Internal method to generate the branch of one node.
	 * 
	 * @param parent
	 *            node to generate branch for
	 * @param nodelist
	 *            list of child pages
	 * 
	 * @throws XPathExpressionException
	 */
	private void generateBranch(TreeNode parent, NodeList nodelist) throws XPathExpressionException {
		// walk through children of the parent...
		for(int n = 0; n < nodelist.getLength(); ++n) {
			Node node = nodelist.item(n); // get current child

			// if page should be shown...
			if(this.viewedPages.indexOf(XMLUtils.getAttributeValue(node, "id")) != -1) {
				String pageID = XMLUtils.getAttributeValue(node, "id");
				String pageName = XMLUtils.getAttributeValue(node, "name");

				if(!this.keyword.equals("") && pageName.toLowerCase().contains(this.keyword)) {
					pageName = "<html><b>" + pageName + "</b></html>";
				}

				// generate menu tree element for current child...
				TreeNode child = new TreeNode(pageName, pageID, node);

				parent.add(child); // append child to parent

				// get children of current child...
				NodeList childNodeList = this.simpleXPath.getNodeSet("page", node);

				// generate branch for the children if there are some...
				if(childNodeList.getLength() > 0) {
					this.generateBranch(child, childNodeList);
				}
			}
		}
	}

	/**
	 * Internal method to create the option panel for one page.
	 * 
	 * @param pageID
	 *            id of the page to generate the option panel for
	 * @param pageName
	 *            name of the page to generate the option panel for
	 * @param defaultValues
	 *            boolean whether or not default values should be used
	 * 
	 * @return the panel for the wanted page
	 * 
	 * @throws XPathExpressionException
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 * @throws NoSuchMethodException
	 * @throws ClassNotFoundException
	 * @throws IllegalArgumentException
	 * @throws SecurityException
	 */
	private JPanel generateOptionPanel(final String pageID, final String pageName, final boolean defaultValues) throws XPathExpressionException, SecurityException, IllegalArgumentException, ClassNotFoundException, NoSuchMethodException, IllegalAccessException, InvocationTargetException {
		JPanel containerPanel = new JPanel(new GridBagLayout());

		GridBagConstraints panelGBC = new GridBagConstraints();
		panelGBC.anchor = GridBagConstraints.NORTH;
		panelGBC.gridx = panelGBC.gridy = 0;
		panelGBC.gridwidth = panelGBC.gridheight = 1;
		panelGBC.weightx = panelGBC.weighty = 1.0;
		panelGBC.fill = GridBagConstraints.BOTH;
		panelGBC.insets = new Insets(10, 10, 10, 10);

		JPanel panel = new JPanel(new GridBagLayout());

		// add page title...
		panel.add(new JLabel("<html><b>" + pageName + "</b></html>"), panelGBC);

		// --- page history - begin ---
		JPanel historyPanel = this.history.addButtons(pageID);

		panelGBC.weightx = 0;
		panelGBC.gridx++;
		panelGBC.anchor = GridBagConstraints.EAST;

		panel.add(historyPanel, panelGBC);
		// --- page history - end ---

		panelGBC.anchor = GridBagConstraints.NORTH;
		panelGBC.gridwidth = 2;
		panelGBC.gridx = 0;
		panelGBC.weightx = 1.0;
		panelGBC.gridy++;
		panelGBC.insets = new Insets(0, 0, 0, 0);

		panel.add(new JSeparator(), panelGBC);

		panelGBC.gridy++;

		// handle options...
		NodeList optionsNodes = this.simpleXPath.getNodeSet("//page[@name='" + pageName + "']/options/child::*", this.menuXML);

		this.parseOptions(pageName, panel, panelGBC, optionsNodes, defaultValues);


		// main panel of the page...
		GridBagConstraints gbc = new GridBagConstraints();
		gbc.anchor = GridBagConstraints.NORTH;
		gbc.gridx = gbc.gridy = 0;
		gbc.gridwidth = gbc.gridheight = 1;
		gbc.weightx = gbc.weighty = 1.0;
		gbc.fill = GridBagConstraints.HORIZONTAL;

		containerPanel.add(panel, gbc);


		// buttons at the bottom of the page...
		JButton restoreButton = new JButton("Restore Defaults");
		restoreButton.setMargin(new Insets(3, 20, 3, 20));
		restoreButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent arg0) {
				// create option panel with default values and add it to list...
				try {
					optionPanels.put(pageID, that.generateOptionPanel(pageID, pageName, true));
				}
				catch(Exception e) {
					e.printStackTrace();
				}

				that.updateRightComponent(pageID);
			}
		});

		JButton applyButton = new JButton("Apply");
		applyButton.setMargin(new Insets(3, 20, 3, 20));
		applyButton.addActionListener(new ActionListener() {
			private String pName = pageName;

			public void actionPerformed(ActionEvent e) {
				try {
					that.savePage(this.pName); // save current page
				}
				catch(Exception e1) {
					e1.printStackTrace();
				}
			}
		});

		JPanel buttonPanel = new JPanel(new FlowLayout(FlowLayout.RIGHT));
		buttonPanel.add(restoreButton);
		buttonPanel.add(applyButton);

		gbc.anchor = GridBagConstraints.SOUTH;
		gbc.gridy++;

		containerPanel.add(buttonPanel, gbc);

		return containerPanel;
	}

	/**
	 * Internal method to parse the options for the given option panel.
	 * 
	 * @param pageName
	 *            the name of the page this option is on
	 * @param optionPanel
	 *            option panel to add parsed options to
	 * @param panelGBC
	 *            GridBagConstraints for the option panel
	 * @param optionsNodes
	 *            NodeList that holds the options for the option panel
	 * @param defaultValues
	 *            Boolean whether or not default values should be used
	 * 
	 * @throws XPathExpressionException
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 * @throws NoSuchMethodException
	 * @throws ClassNotFoundException
	 * @throws IllegalArgumentException
	 * @throws SecurityException
	 */
	private void parseOptions(String pageName, JPanel optionPanel, GridBagConstraints panelGBC, NodeList optionsNodes, boolean defaultValues) throws XPathExpressionException, SecurityException, IllegalArgumentException, ClassNotFoundException, NoSuchMethodException, IllegalAccessException, InvocationTargetException {
		// walk through options...
		for(int i = 0; i < optionsNodes.getLength(); i += 1) {
			Node optionNode = optionsNodes.item(i); // get node of current option

			// --- handle separator lines - begin ---
			if(optionNode.getNodeName().equals("line")) {
				panelGBC.insets = new Insets(0, 10, 0, 10);

				optionPanel.add(new JSeparator(), panelGBC);

				panelGBC.gridy++;
				panelGBC.insets = new Insets(0, 0, 0, 0);

				continue;
			}
			// --- handle separator lines - end ---

			// --- handle text - begin ---
			if(optionNode.getNodeName().equals("text")) {
				panelGBC.insets = new Insets(0, 15, 0, 15);

				String text = this.simpleXPath.getString("./text()", optionNode);

				optionPanel.add(new JLabel(text), panelGBC);

				panelGBC.gridy++;
				panelGBC.insets = new Insets(0, 0, 0, 0);

				continue;
			}
			// --- handle text - end ---

			// --- handle visibility - begin ---
			String visible = XMLUtils.getAttributeValue(optionNode, "visible");

			if(visible != null && !visible.equals("") && !visible.equalsIgnoreCase("gui")) {
				continue;
			}
			// --- handle visibility - end ---

			String value = XMLUtils.getAttributeValue(optionNode, "value");
			String nodeName = XMLUtils.getAttributeValue(optionNode, "name");
			String helpText = XMLUtils.getAttributeValue(optionNode, "help");

			// handle option boxes...
			if(optionNode.getNodeName().equals("optionbox")) {
				JPanel boxPanel = new JPanel(new GridBagLayout());
				boxPanel.setBorder(new TitledBorder(new LineBorder(Color.BLACK), nodeName));

				GridBagConstraints boxPanelGBC = new GridBagConstraints();
				boxPanelGBC.anchor = GridBagConstraints.NORTH;
				boxPanelGBC.gridx = boxPanelGBC.gridy = 0;
				boxPanelGBC.gridwidth = boxPanelGBC.gridheight = 1;
				boxPanelGBC.weightx = boxPanelGBC.weighty = 1.0;
				boxPanelGBC.fill = GridBagConstraints.HORIZONTAL;

				NodeList optionsNodeList = this.simpleXPath.getNodeSet("./child::*", optionNode);

				this.parseOptions(pageName, boxPanel, boxPanelGBC, optionsNodeList, defaultValues);

				this.handleDependsOnAttribute(pageName, "", optionNode, boxPanel);

				panelGBC.insets = new Insets(0, 10, 0, 10);

				optionPanel.add(boxPanel, panelGBC);

				panelGBC.gridy++;
				panelGBC.insets = new Insets(0, 0, 0, 0);

				continue;
			}

			// create row panel for the current option...
			JPanel rowPanel = new JPanel(new GridBagLayout());

			GridBagConstraints gbc = new GridBagConstraints();
			gbc.anchor = GridBagConstraints.WEST;
			gbc.gridx = gbc.gridy = 0;
			gbc.gridwidth = gbc.gridheight = 1;
			gbc.weightx = gbc.weighty = 1.0;
			gbc.insets = new Insets(5, 15, 5, 5);


			// create label for the option...
			JLabel optionLabel = new JLabel(nodeName);

			if(helpText != null) {
				optionLabel.setToolTipText(helpText);
			}

			this.handleDependsOnAttribute(pageName, "", optionNode, optionLabel);

			rowPanel.add(optionLabel, gbc); // add label to row panel


			// get value panel...
			JPanel valuePanel = this.parseOptionValue(optionNode, pageName, value, 0, defaultValues);

			this.handleDependsOnAttribute(pageName, "", optionNode, valuePanel);


			gbc.anchor = GridBagConstraints.EAST;
			gbc.gridx++;
			gbc.insets = new Insets(5, 5, 5, 15);

			rowPanel.add(valuePanel, gbc);

			panelGBC.fill = GridBagConstraints.BOTH;

			optionPanel.add(rowPanel, panelGBC);

			panelGBC.gridy++;
		}
	}

	/**
	 * Internal method to parse an option value which consider lists and
	 * sequences.
	 * 
	 * @param optionNode
	 *            the current option node
	 * @param pageName
	 *            the name of the page this option is on
	 * @param value
	 *            string representation of the value of this option
	 * @param index
	 *            index to parse from the option list
	 * @param defaultValues
	 *            Boolean whether or not default values should be used
	 * 
	 * @return a JPanel with the visual representation of the value-string on it
	 * 
	 * @throws ClassNotFoundException
	 * @throws NoSuchMethodException
	 * @throws SecurityException
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 * @throws IllegalArgumentException
	 * @throws XPathExpressionException
	 */
	private JPanel parseOptionValue(final Node optionNode, final String pageName, String value, int index, final boolean defaultValues) throws ClassNotFoundException, SecurityException, NoSuchMethodException, IllegalArgumentException, IllegalAccessException, InvocationTargetException, XPathExpressionException {
		// get ID, node name and data type of the option...
		final String id = XMLUtils.getAttributeValue(optionNode, "id");
		String nodeName = XMLUtils.getAttributeValue(optionNode, "name");
		String datatype = XMLUtils.getAttributeValue(optionNode, "datatype");

		// create the pattern to match and create a matcher against the string...
		Matcher matcher = null;

		if(value != null) {
			Pattern pattern = Pattern.compile("\\{\\{(.*)\\}\\}");
			matcher = pattern.matcher(value);
		}

		if(value != null && matcher.find() == true) { // option has a list of values...
			JPanel listPanel = new JPanel(new GridBagLayout());

			GridBagConstraints listPanelGBC = new GridBagConstraints();
			listPanelGBC.anchor = GridBagConstraints.WEST;
			listPanelGBC.gridx = listPanelGBC.gridy = 0;
			listPanelGBC.gridwidth = listPanelGBC.gridheight = 1;
			listPanelGBC.weightx = listPanelGBC.weighty = 1.0;
			listPanelGBC.insets = new Insets(0, 0, 0, 0);


			final JPanel valuePanel = new JPanel(new GridBagLayout()); // create empty panel

			final GridBagConstraints listValueGBC = new GridBagConstraints();
			listValueGBC.anchor = GridBagConstraints.WEST;
			listValueGBC.gridx = listValueGBC.gridy = 0;
			listValueGBC.gridwidth = listValueGBC.gridheight = 1;
			listValueGBC.weightx = listValueGBC.weighty = 1.0;
			listValueGBC.insets = new Insets(5, 0, 5, 0);


			// get elements of the list...
			String[] listElements = matcher.group(1).split(" \\$\\$ ");

			// get minimum amount...
			int minimum = 1;

			String minString = XMLUtils.getAttributeValue(optionNode, "minimum");

			if(minString != null) { // minimum provided...
				minimum = Integer.parseInt(minString);
			}

			// check lower bound...
			if(listElements.length < minimum) { // the lower limit is higher then the amount of elements we have...
				int missing = minimum - listElements.length; // amount of missing elements

				// create new list which is big enough...
				String[] newListElements = new String[listElements.length + missing];

				// copy elements from the old list into the new...
				for(int i = 0; i < listElements.length; i += 1) {
					newListElements[i] = listElements[i];
				}

				Class<?> c = Class.forName(datatype); // get class

				Object obj = null;

				Method m = c.getMethod("serializeElement", Object.class); // get method

				// add amount of missing elements...
				for(int i = 0; i < missing; i += 1) {
					newListElements[listElements.length] = (String) m.invoke(c, obj); // invoke method
				}

				// write new list back into variable of the old list...
				listElements = newListElements;
			}


			// get maximum amount...
			int maximum = -1;

			String maxString = XMLUtils.getAttributeValue(optionNode, "maximum");

			if(maxString != null) { // maximum provided...
				maximum = Integer.parseInt(maxString);
			}


			// final minimum and maximum variables...
			final int minValue = minimum;
			final int maxValue = maximum;


			// check upper bound...
			if(maximum == -1 || maximum > listElements.length) { // no upper limit or upper limit is higher then amount of elements...
				maximum = listElements.length;
			}


			// define plus button...
			final JButton plusButton = new JButton("+");
			plusButton.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent ae) {
					// get amount of elements...
					int count = getChangeListAmount(pageName, id);

					// add element...
					try {
						valuePanel.add(addListElement(listValueGBC, that.parseOptionValue(optionNode, pageName, null, count, false), pageName, id, count, minValue, maxValue));
					}
					catch(Exception e) {
						e.printStackTrace();
					}

					count += 1; // increment amount of elements by one

					// if maximum is reached...
					if(count == maxValue) {
						plusButton.setEnabled(false);
					}

					// update GUI...
					updateRightComponent(pageName);
				}
			});

			this.plusButtonList.put(id, plusButton); // add button to list


			// add list elements...
			for(int i = 0; i < maximum; i += 1) {
				valuePanel.add(this.addListElement(listValueGBC, this.parseOptionValue(optionNode, pageName, listElements[i], i, defaultValues), pageName, id, i, minValue, maxValue));
			}

			listPanel.add(valuePanel, listPanelGBC);
			listPanelGBC.gridx++;
			listPanelGBC.insets = new Insets(5, 15, 5, 0);


			listPanel.add(plusButton, listPanelGBC);


			return listPanel;
		}

		if(this.sequencesList.containsKey(datatype)) { // option is a sequence...
			JPanel valuePanel = new JPanel(new GridBagLayout()); // create empty panel

			GridBagConstraints sequenceGBC = new GridBagConstraints();
			sequenceGBC.anchor = GridBagConstraints.WEST;
			sequenceGBC.gridx = sequenceGBC.gridy = 0;
			sequenceGBC.gridwidth = sequenceGBC.gridheight = 1;
			sequenceGBC.weightx = sequenceGBC.weighty = 1.0;
			sequenceGBC.insets = new Insets(5, 15, 5, 5);

			// get sequence options...
			NodeList sequenceOptions = this.simpleXPath.getNodeSet("./option", this.sequencesList.get(datatype));

			// walk through sequence options...
			for(int j = 0; j < sequenceOptions.getLength(); j += 1) {
				// get current sequence option node...
				Node sequenceOption = sequenceOptions.item(j);

				// get ID and data type for this sequence option node...
				String nodeID = id + "." + XMLUtils.getAttributeValue(sequenceOption, "id");
				nodeName = XMLUtils.getAttributeValue(sequenceOption, "name");
				datatype = XMLUtils.getAttributeValue(sequenceOption, "datatype");

				String helpText = XMLUtils.getAttributeValue(sequenceOption, "help");

				JLabel sequenceLabel = new JLabel(nodeName);

				if(helpText != null) {
					sequenceLabel.setToolTipText(helpText);
				}

				this.handleDependsOnAttribute(pageName, id + ".", sequenceOption, sequenceLabel);

				valuePanel.add(sequenceLabel, sequenceGBC);


				// manipulate the GBC for the next element...
				sequenceGBC.gridx++;
				sequenceGBC.insets = new Insets(5, 5, 5, 0);

				// get value panel for this sequence option...
				JPanel optionValuePanel = this.createOptionValuePanel(this.xPref.getValues(nodeID, defaultValues).get(index), nodeID, datatype, sequenceOption, pageName, index);

				this.handleDependsOnAttribute(pageName, id + ".", sequenceOption, optionValuePanel);

				// add panel for this sequence option to the value panel...
				valuePanel.add(optionValuePanel, sequenceGBC);

				// manipulate the GBC for the next element...
				sequenceGBC.gridx++;
			}

			// walk through sequence options...
			for(int j = 0; j < sequenceOptions.getLength(); j += 1) {
				// get current sequence option node...
				Node sequenceOption = sequenceOptions.item(j);

				// get ID for this sequence option node...
				String nodeID = id + "." + XMLUtils.getAttributeValue(sequenceOption, "id");

				this.statusOfDependencyChanged(nodeID, this.changeList.get(pageName).get(nodeID).get(index));
			}

			return valuePanel;
		}

		// option is a normal value...
		// get value panel for this option...
		Object obj = null;

		if(value != null) {
			obj = this.xPref.getValues(id, defaultValues).get(index);
		}

		return this.createOptionValuePanel(obj, id, datatype, optionNode, pageName, index);
	}

	/**
	 * Internal method to add the GUI representation of a new list element
	 * 
	 * @param listGBC
	 *            the GridBagConstraints of the panels
	 * @param valuePanel
	 *            the valuePanel to add this representation for
	 * 
	 * @return the JPanel with this list element
	 */
	private JPanel addListElement(GridBagConstraints listGBC, JPanel valuePanel, final String pageName, final String id, final Integer index, final int minimum, final int maximum) {
		final JPanel panel = new JPanel(new GridBagLayout());

		// get value panel for this sequence option and add it the value panel...
		panel.add(valuePanel, listGBC);
		listGBC.gridx++;

		final JButton minusButton = new JButton("-");
		minusButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				// remove panel...
				panel.getParent().remove(panel);

				// update component...
				updateRightComponent(pageName);

				// remove the change value...
				valueChanged(pageName, id, null, index);

				// remove button from list...
				minusButtonList.get(id).remove(minusButton);


				// get amount of elements...
				int count = getChangeListAmount(pageName, id);

				// count has dropped below maximum...
				if(count < maximum) {
					plusButtonList.get(id).setEnabled(true); // enable plus button for this id
				}

				// minimum reached...
				if(count == minimum) {
					setMinusButtonState(id, false); // disable minus buttons for this id
				}
			}
		});

		// add button to list...
		if(!this.minusButtonList.containsKey(id)) {
			this.minusButtonList.put(id, new LinkedList<JButton>());
		}

		this.minusButtonList.get(id).add(minusButton);


		// get amount of elements...
		int count = this.getChangeListAmount(pageName, id);

		// maximum reached...
		if(count == maximum) {
			this.plusButtonList.get(id).setEnabled(false); // disable plus button for this id
		}


		panel.add(minusButton, listGBC);
		listGBC.gridx++;
		listGBC.insets = new Insets(5, 15, 5, 5);
		listGBC.fill = GridBagConstraints.VERTICAL;

		panel.add(new JSeparator(SwingConstants.VERTICAL), listGBC);
		listGBC.gridx++;
		listGBC.insets = new Insets(5, 0, 5, 0);
		listGBC.fill = GridBagConstraints.NONE;

		return panel;
	}

	/**
	 * Internal method to set the state of all minus buttons of one option...
	 * 
	 * @param id
	 *            id of the option
	 * @param newState
	 *            new state
	 */
	private void setMinusButtonState(String id, boolean newState) {
		for(JButton minusButton : this.minusButtonList.get(id)) {
			minusButton.setEnabled(newState);
		}
	}

	/**
	 * Internal method to get the amount of used elements in the change list for
	 * the given id.
	 * 
	 * @param pageName
	 *            name of the page we are on
	 * @param id
	 *            id we want to know the amount from
	 * 
	 * @return amount of used elements for this id
	 */
	private int getChangeListAmount(String pageName, String id) {
		int count = 0;

		for(String element : this.changeList.get(pageName).get(id)) {
			if(element != null) {
				count += 1;
			}
		}

		return count;
	}

	/**
	 * Internal method to create the panel for one option value.
	 * 
	 * @param id
	 *            the id this panel is for
	 * @param datatype
	 *            the data type of the value to be represented by this panel
	 * @param optionNode
	 *            the optionNode of this id
	 * @param pageName
	 *            the name of the page this panel will be on
	 * @param index
	 *            index of the values list for this id
	 * 
	 * @return the value panel
	 * 
	 * @throws ClassNotFoundException
	 * @throws NoSuchMethodException
	 * @throws SecurityException
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 * @throws IllegalArgumentException
	 */
	private JPanel createOptionValuePanel(Object obj, String id, String datatype, Node optionNode, String pageName, Integer index) throws ClassNotFoundException, SecurityException, NoSuchMethodException, IllegalArgumentException, IllegalAccessException, InvocationTargetException {
		JPanel panel = new JPanel();

		// get panel...
		Class<?> c = Class.forName(datatype); // get class

		Method m = c.getMethod("getVisualRepresentation", String.class, Object.class, Node.class, XPrefGUI.class, String.class, Integer.class); // get method

		panel = (JPanel) m.invoke(c, id, obj, optionNode, this, pageName, index); // invoke method

		// set change value...
		m = c.getMethod("serializeElement", Object.class);

		String value = (String) m.invoke(c, obj);

		this.valueChanged(pageName, id, value, index);

		return panel;
	}

	/**
	 * Internal Method to set a panel as right component of the window.
	 * 
	 * @param pageID
	 *            name of the current page
	 */
	private void updateRightComponent(String pageID) {
		JPanel pagePanel = this.optionPanels.get(pageID);

		int dividerLocation = this.mainSP.getDividerLocation();

		JScrollPane panelScrollPane = new JScrollPane(pagePanel);
		panelScrollPane.getVerticalScrollBar().setUnitIncrement(10);
		panelScrollPane.getHorizontalScrollBar().setUnitIncrement(10);

		// add panel to SplitPane...
		this.mainSP.setRightComponent(panelScrollPane);
		this.mainSP.setDividerLocation(dividerLocation);
	}


	/**
	 * ========== semi public methods ==========
	 */

	/**
	 * Add the new value to the change list.
	 * 
	 * @param pageName
	 *            name of the page this value is on
	 * @param id
	 *            id of the element this value belongs to
	 * @param value
	 *            new value
	 * @param index
	 *            the position of this new value in the list
	 */
	public void valueChanged(String pageName, String id, String value, int index) {
		this.secure();

		// create changeList for named page if it does not exist...
		if(!this.changeList.containsKey(pageName)) {
			this.changeList.put(pageName, new Hashtable<String, LinkedList<String>>());
		}

		Hashtable<String, LinkedList<String>> pageChangeList = this.changeList.get(pageName);

		// create a list for the named id if it does not exist....
		if(!pageChangeList.containsKey(id)) {
			pageChangeList.put(id, new LinkedList<String>());
		}

		LinkedList<String> idChangeList = pageChangeList.get(id);

		// add value to named page of changeList...
		if(idChangeList.size() > index) {
			idChangeList.set(index, value);
		}
		else {
			idChangeList.add(value);
		}


		// handle dependency changes...
		this.statusOfDependencyChanged(id, value);
	}

	/**
	 * Internal method to process the status change of a dependency.
	 * 
	 * @param nodeID
	 *            the ID of the element which status changed
	 * @param newValue
	 *            the new value
	 */
	private void statusOfDependencyChanged(String nodeID, String newValue) {
		// something depends on a value of this id?...
		if(this.dependencyList.containsKey(nodeID)) {
			// disable all elements that have dependencies to this id...
			for(LinkedList<JComponent> valueList : this.dependencyList.get(nodeID).values()) {
				for(JComponent component : valueList) {
					this.changeComponentState(component, false);
				}
			}

			// are there any dependencies for the new value?...
			if(this.dependencyList.get(nodeID).containsKey(newValue)) {
				// enable the elements that depend on this value...
				for(JComponent component : this.dependencyList.get(nodeID).get(newValue)) {
					this.changeComponentState(component, true);
				}
			}
		}
	}

	/**
	 * Internal method to handle the dependsOn attribute and add the
	 * dependencies to the data structure, if there are any.
	 * 
	 * @param pageName
	 * @param nodeIDPrefix
	 *            prefix of the node id (needed for sequences)
	 * @param optionNode
	 *            option node of the element that depends on an other
	 * @param component
	 *            component that is the visual representation of the option
	 */
	private void handleDependsOnAttribute(String pageName, String nodeIDPrefix, Node optionNode, JComponent component) {
		String dependsOnString = XMLUtils.getAttributeValue(optionNode, "dependsOn");

		// there is no dependsOn attribute...
		if(dependsOnString == null) {
			return;
		}

		// split dependsOn attribute to get the id and the true values...
		String[] dependsOnElements = dependsOnString.split("=");

		String nodeID = nodeIDPrefix + dependsOnElements[0];
		String[] trueValues = dependsOnElements[1].split("\\|");

		// no dependency for this nodeID yet...
		if(!this.dependencyList.containsKey(nodeID)) {
			this.dependencyList.put(nodeID, new Hashtable<String, LinkedList<JComponent>>());
		}

		// walk through list of true values...
		for(String trueValue : trueValues) {
			// no dependency for this value yet...
			if(!this.dependencyList.get(nodeID).containsKey(trueValue)) {
				this.dependencyList.get(nodeID).put(trueValue, new LinkedList<JComponent>());
			}

			// add dependency for panel...
			this.dependencyList.get(nodeID).get(trueValue).add(component);
		}

		this.statusOfDependencyChanged(nodeID, this.changeList.get(pageName).get(nodeID).get(0));
	}

	/**
	 * Internal method to notify all registered components that the preferences
	 * have changed.
	 */
	@SuppressWarnings("unchecked")
	private synchronized void notifyComponents() {
		for(IXPref component : (Iterable<IXPref>) this.componentsList.clone()) {
			component.preferencesChanged();
		}


		SwingUtilities.updateComponentTreeUI(XPrefGUI.this);

		for(JPanel panel : this.optionPanels.values()) {
			SwingUtilities.updateComponentTreeUI(panel);
		}
	}

	/**
	 * Internal method to save the changes of one page.
	 * 
	 * @param pageName
	 *            Name of the page to save changes for
	 * 
	 * @throws TransformerException
	 * @throws IOException
	 * @throws ParserConfigurationException
	 * @throws XPathExpressionException
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 * @throws NoSuchMethodException
	 * @throws ClassNotFoundException
	 * @throws IllegalArgumentException
	 * @throws SecurityException
	 */
	private void savePage(String pageName) throws IOException, TransformerException, ParserConfigurationException, XPathExpressionException, SecurityException, IllegalArgumentException, ClassNotFoundException, NoSuchMethodException, IllegalAccessException, InvocationTargetException {
		// we can access the file system and have a name for save XML...
		if(this.accessToFileSystem && this.saveXMLFileName != null) {
			// initialize the save XML...
			this.initializeSaveXML();

			// save the data of the current page...
			this.saveData(pageName);

			// write the created save XML to the file system...
			XMLUtils.writeXMLFile(this.saveXMLFileName, this.saveXML);
		}

		// apply changes to the running system...
		this.applyChanges();
	}

	/**
	 * Internal method to save the changes of all pages.
	 * 
	 * @throws TransformerException
	 * @throws IOException
	 * @throws ParserConfigurationException
	 * @throws XPathExpressionException
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 * @throws NoSuchMethodException
	 * @throws ClassNotFoundException
	 * @throws IllegalArgumentException
	 * @throws SecurityException
	 */
	private void saveAllPages() throws IOException, TransformerException, ParserConfigurationException, XPathExpressionException, SecurityException, IllegalArgumentException, ClassNotFoundException, NoSuchMethodException, IllegalAccessException, InvocationTargetException {
		// we can access the file system and have a name for save XML...
		if(this.accessToFileSystem && this.saveXMLFileName != null) {
			// initialize the save XML...
			this.initializeSaveXML();

			// walk through all pages...
			for(String pageName : this.changeList.keySet()) {
				// save the data of the current page...
				this.saveData(pageName);
			}

			// write the created save XML to the file system...
			XMLUtils.writeXMLFile(this.saveXMLFileName, this.saveXML);
		}

		// apply changes to the running system...
		this.applyChanges();
	}

	/**
	 * Internal method to apply all changes to the running system.
	 * 
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 * @throws NoSuchMethodException
	 * @throws ClassNotFoundException
	 * @throws IllegalArgumentException
	 * @throws SecurityException
	 * @throws XPathExpressionException
	 */
	private void applyChanges() throws XPathExpressionException, SecurityException, IllegalArgumentException, ClassNotFoundException, NoSuchMethodException, IllegalAccessException, InvocationTargetException {
		// walk through all pages...
		for(String pageName : this.changeList.keySet()) {
			// get the changes for this page...
			Hashtable<String, LinkedList<String>> changeList = this.changeList.get(pageName);

			// walk through the ID's of the change set,
			// concatenate the elements into a list value string,
			// parse the list value string as a saved value...
			for(String id : changeList.keySet()) {
				this.xPref.valuesList.remove(id);

				LinkedList<String> valuesList = changeList.get(id);

				StringBuffer valuesString = new StringBuffer();

				if(valuesList.size() == 1) {
					valuesString.append(valuesList.get(0));
				}
				else {
					valuesString.append("{{");

					for(int i = 0; i < valuesList.size(); i += 1) {
						valuesString.append(valuesList.get(i));

						if(i < valuesList.size()-1) {
							valuesString.append(" $$ ");
						}
					}

					valuesString.append("}}");
				}

				this.xPref.parseSavedValue(id, valuesString.toString(), false);
			}
		}

		// notify the registered components that there are changes...
		this.notifyComponents();
	}

	/**
	 * Internal method to initialize the DOMDocument for the save XML if it does
	 * not exist.
	 * 
	 * @throws ParserConfigurationException
	 */
	private void initializeSaveXML() throws ParserConfigurationException {
		if(this.saveXML == null) {
			DocumentBuilderFactory documentBuilderFactory = DocumentBuilderFactory.newInstance();
			DocumentBuilder documentBuilder = documentBuilderFactory.newDocumentBuilder();

			this.saveXML = documentBuilder.newDocument();
			this.saveXML.appendChild(this.saveXML.createElement("savedPreferences"));
		}
	}

	/**
	 * Saves the given partial saveList into the saveXML.
	 *
	 * @param pageName
	 *            name of page to save
	 * @param saveList
	 *            saveList of the page
	 * 
	 * @throws XPathExpressionException
	 */
	private void saveData(String pageName) throws XPathExpressionException {
		this.removePageOptionsFromSaveXML(pageName);

		Hashtable<String, LinkedList<String>> saveList = this.getSaveList(pageName);

		// stop, if there is no saveList...
		if(saveList == null) {
			return;
		}

		// get rootNode of saveXML...
		Node rootNode = this.saveXML.getDocumentElement();

		// walk through saveList and create Elements for the children in
		// saveXML...
		for(String id : saveList.keySet()) {
			LinkedList<String> values = saveList.get(id);

			StringBuffer value = new StringBuffer();

			if(values.size() == 1) {
				value.append(values.get(0));
			}
			else {
				value.append("{{");

				for(int i = 0; i < values.size(); i += 1) {
					value.append(values.get(i));

					if(i < values.size()-1) {
						value.append(" $$ ");
					}
				}

				value.append("}}");
			}

			Element optionNode = this.saveXML.createElement("item");
			optionNode.setAttribute("id", id);
			optionNode.appendChild(this.saveXML.createTextNode(value.toString()));

			rootNode.appendChild(optionNode);
		}
	}

	/**
	 * Internal method to remove the default values from the change list.
	 * 
	 * @param pageName
	 *            the name of the page for which to do this
	 * 
	 * @return the save list
	 */
	private Hashtable<String, LinkedList<String>> getSaveList(String pageName) {
		Hashtable<String, LinkedList<String>> saveList = this.changeList.get(pageName);

		if(saveList == null) {
			return null;
		}

		Hashtable<String, LinkedList<String>> defaultList = this.defaultValueList.get(pageName);

		for(String id : defaultList.keySet()) {
			if(saveList.containsKey(id)) {
				boolean changed = false;

				for(String savedValue : saveList.get(id)) {
					if(!defaultList.get(id).contains(savedValue)) {
						changed = true;

						break;
					}
				}

				if(!changed) {
					saveList.remove(id);
				}
			}
		}

		return saveList;
	}

	/**
	 * Removes all settings of the given page from the saveXML.
	 * 
	 * @param pageName
	 *            page name
	 * 
	 * @throws XPathExpressionException
	 */
	private void removePageOptionsFromSaveXML(String pageName) throws XPathExpressionException {
		// get option items of this page...
		NodeList pageOptions = this.simpleXPath.getNodeSet("//page[@name='" + pageName + "']/options//option", this.menuXML);

		// walk through option items of this page...
		for(int i = 0; i < pageOptions.getLength(); ++i) {
			// get current option...
			Node pageOption = pageOptions.item(i);

			// get id of current option...
			String id = XMLUtils.getAttributeValue(pageOption, "id");
			String datatype = XMLUtils.getAttributeValue(pageOption, "datatype");

			if(this.sequencesList.containsKey(datatype)) {
				NodeList sequenceNodeList = this.simpleXPath.getNodeSet("/preferences/sequence[@name='" + datatype + "']/option", this.menuXML);

				for(int n = 0; n < sequenceNodeList.getLength(); ++n) {
					Node innerNode = sequenceNodeList.item(n);

					String innerID = id + "." + XMLUtils.getAttributeValue(innerNode, "id");

					this.removeIDFromSaveXML(innerID);
				}
			}
			else {
				this.removeIDFromSaveXML(id);
			}
		}
	}

	private void removeIDFromSaveXML(String id) throws XPathExpressionException {
		// get saved option node from saveXML...
		Node optionNode = this.simpleXPath.getNode("/savedPreferences/item[@id='" + id + "']", this.saveXML);

		if(optionNode != null) { // if entry in saveXML exists...
			//			optionNode.getParentNode().removeChild(optionNode.getNextSibling()); // delete new line node after the current one
			optionNode.getParentNode().removeChild(optionNode); // delete it
		}
	}
}