package xpref.datatypes;

import java.awt.Color;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.io.IOException;
import java.lang.instrument.IllegalClassFormatException;
import java.lang.reflect.InvocationTargetException;
import java.util.LinkedList;
import java.util.regex.Pattern;

import javax.swing.JButton;
import javax.swing.JColorChooser;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.xpath.XPathExpressionException;

import org.w3c.dom.Node;
import org.xml.sax.SAXException;

import xpref.XPref;
import xpref.XPrefGUI;

public class ColorDatatype {
	/**
	 * The default value for this data type.
	 */
	protected static Color defaultValue = Color.BLACK;

	/**
	 * Static method to serialize the given element.
	 * 
	 * @param element
	 *            the element to serialize
	 * 
	 * @return the serialized element as string
	 */
	public static String serializeElement(Object element) {
		if(element == null) {
			element = ColorDatatype.defaultValue;
		}

		return "#" + Integer.toHexString(((Color) element).getRGB()).substring(2).toUpperCase();
	}

	/**
	 * Static method to parse the string representation of the element to the
	 * right data type.
	 * 
	 * @param optionNode
	 *            the option node to parse the element for
	 * @param element
	 *            the string representation of the element
	 * 
	 * @return the element as instance of the right data type
	 */
	public static Color parseElement(Node optionNode, String element) {
		if(element == null) {
			return ColorDatatype.defaultValue;
		}
		else {
			String colString = element.substring(1);

			int r = Integer.parseInt(colString.substring(0, 2), 16);
			int g = Integer.parseInt(colString.substring(2, 4), 16);
			int b = Integer.parseInt(colString.substring(4, 6), 16);

			return new Color(r, g, b);
		}
	}

	/**
	 * Static method to get the visual representation of the element for
	 * XPrefGUI.
	 * 
	 * @param element
	 *            the element to represent
	 * @param optionNode
	 *            node of the element
	 * @param xPrefGUI
	 *            instance of the XPrefGUI
	 * @param pageName
	 *            name of the page this value is on
	 * 
	 * @return the panel holding the visual representation
	 */
	public static JPanel getVisualRepresentation(final String nodeID, Object element, Node optionNode, final XPrefGUI xPrefGUI, final String pageName, final Integer index) {
		final JTextField colorTF = new JTextField(ColorDatatype.serializeElement(element));
		colorTF.addFocusListener(new FocusListener() {
			public void focusGained(FocusEvent fe) {}

			public void focusLost(FocusEvent fe) {
				String saveValue = ((JTextField) fe.getComponent()).getText();

				if(!Pattern.matches("^#[A-Fa-f0-9]{6}$", saveValue)) {
					JOptionPane.showMessageDialog(colorTF, "Malformed color value!", "Error", JOptionPane.ERROR_MESSAGE);
					colorTF.grabFocus();
				}

				xPrefGUI.valueChanged(pageName, nodeID, saveValue, index);
			}
		});

		JPanel optionElement = new JPanel();
		optionElement.add(colorTF);

		final JButton button = new JButton("choose");
		button.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				Color newColor = JColorChooser.showDialog(button, "Choose Color", ColorDatatype.parseElement(null, colorTF.getText()));

				if(newColor != null) {
					String colorString = ColorDatatype.serializeElement(newColor);

					colorTF.setText(colorString);

					xPrefGUI.valueChanged(pageName, nodeID, colorString, index);
				}
			}
		});

		optionElement.add(button);

		// wrap check box in a panel...
		JPanel panel = new JPanel();
		panel.add(optionElement);

		return panel;
	}

	/**
	 * Static method to get the first value of the given id as instance of the
	 * proper data type.
	 * 
	 * @param id
	 *            the id to get the value for
	 * 
	 * @return the element as instance of the right data type
	 * 
	 * @throws IOException
	 * @throws ParserConfigurationException
	 * @throws SAXException
	 * @throws XPathExpressionException
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 * @throws IllegalClassFormatException
	 * @throws NoSuchMethodException
	 * @throws NoSuchFieldException
	 * @throws ClassNotFoundException
	 * @throws IllegalArgumentException
	 * @throws SecurityException
	 */
	public static Color getFirstValue(String id) throws XPathExpressionException, SAXException, ParserConfigurationException, IOException, SecurityException, IllegalArgumentException, ClassNotFoundException, NoSuchFieldException, NoSuchMethodException, IllegalClassFormatException, IllegalAccessException, InvocationTargetException {
		LinkedList<Color> list = ColorDatatype.getValues(id);

		return list.get(0);
	}

	/**
	 * Static method to get the values to the given id as instances of the
	 * proper data type.
	 * 
	 * @param id
	 *            the id to get the values for
	 * 
	 * @return the elements as list of instances of the right data type
	 * 
	 * @throws IOException
	 * @throws ParserConfigurationException
	 * @throws SAXException
	 * @throws XPathExpressionException
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 * @throws IllegalClassFormatException
	 * @throws NoSuchMethodException
	 * @throws NoSuchFieldException
	 * @throws ClassNotFoundException
	 * @throws IllegalArgumentException
	 * @throws SecurityException
	 */
	public static LinkedList<Color> getValues(String id) throws XPathExpressionException, SAXException, ParserConfigurationException, IOException, SecurityException, IllegalArgumentException, ClassNotFoundException, NoSuchFieldException, NoSuchMethodException, IllegalClassFormatException, IllegalAccessException, InvocationTargetException {
		LinkedList<Color> colorList = new LinkedList<Color>();

		for(Object obj : XPref.getInstance().getValues(id)) {
			colorList.add((Color) obj);
		}

		return colorList;
	}

	/**
	 * Static method to describe the data type.
	 * 
	 * @return Description of the data type.
	 */
	public static String description() {
		return "Hexadecimal color representation";
	}

	/**
	 * Static method to describe the usage for this option.
	 * 
	 * @param optionNode
	 *            the option node to show the description for
	 * 
	 * @return the description string
	 */
	public static String usage(Node optionNode) {
		return "e.g. (red): #FF0000";
	}

	/**
	 * Static method to get the name for this data type. The name can depend on
	 * some properties of the node it is used for.
	 * 
	 * @param optionNode
	 *            the node this data type is used for
	 * 
	 * @return the name of the data type
	 */
	public static String getName(Node optionNode) {
		return BooleanDatatype.class.getName();
	}
}