package xpref.datatypes;

import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.IOException;
import java.lang.instrument.IllegalClassFormatException;
import java.lang.reflect.InvocationTargetException;
import java.util.LinkedList;

import javax.swing.ButtonGroup;
import javax.swing.JComboBox;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.xpath.XPathExpressionException;

import org.w3c.dom.Node;
import org.xml.sax.SAXException;

import xpref.XPref;
import xpref.XPrefGUI;
import xpref.util.SimpleXPath;
import xpref.util.XMLUtils;

public class EnumDatatype {
	/**
	 * The default value for this data type.
	 */
	protected static String defaultValue = "";

	/**
	 * Static method to serialize the given element.
	 * 
	 * @param element
	 *            the element to serialize
	 * 
	 * @return the serialized element as string
	 */
	public static String serializeElement(Object element) {
		if(element == null) {
			element = EnumDatatype.defaultValue;
		}

		return (String) element;
	}

	/**
	 * Static method to parse the string representation of the element to the
	 * right data type.
	 * 
	 * @param optionNode
	 *            the option node to parse the element for
	 * @param element
	 *            the string representation of the element
	 * 
	 * @return the element as instance of the right data type
	 * 
	 * @throws ClassNotFoundException
	 *             if the class for the data type is not found
	 * @throws NoSuchFieldException
	 *             if the element is not valid
	 */
	public static String parseElement(Node optionNode, String element) throws ClassNotFoundException, NoSuchFieldException {
		if(element == null) {
			return EnumDatatype.defaultValue;
		}
		else {
			LinkedList<String> coBoElements = EnumDatatype.getEnum(optionNode);

			if(coBoElements.contains(element)) {
				return element;
			}
			else {
				throw new NoSuchFieldException("The Enum for the id '" + XMLUtils.getAttributeValue(optionNode, "id") + "' has no member '" + element + "'!");
			}
		}
	}

	/**
	 * Static method to get the visual representation of the element for
	 * XPrefGUI.
	 * 
	 * @param element
	 *            the element to represent
	 * @param optionNode
	 *            node of the element
	 * @param xPrefGUI
	 *            instance of the XPrefGUI
	 * @param pageName
	 *            name of the page this value is on
	 * 
	 * @return the panel holding the visual representation
	 * 
	 * @throws ClassNotFoundException
	 *             if the class for the data type is not found
	 */
	public static JPanel getVisualRepresentation(final String nodeID, Object element, Node optionNode, final XPrefGUI xPrefGUI, final String pageName, final Integer index) throws ClassNotFoundException {
		// wrap check box in a panel...
		JPanel panel = new JPanel();

		LinkedList<String> coBoElements = EnumDatatype.getEnum(optionNode);


		String elem = null;

		if(element == null) {
			elem = EnumDatatype.defaultValue;
		}
		else {
			elem = (String) element;
		}


		int selectIndex = coBoElements.indexOf(elem);

		if(selectIndex == -1) {
			selectIndex = coBoElements.indexOf(XMLUtils.getAttributeValue(optionNode, "value"));
		}

		if(selectIndex == -1) {
			selectIndex = 0;
		}


		String type = XMLUtils.getAttributeValue(optionNode, "parameters");

		// radio buttons...
		if(type != null && type.startsWith("radiobuttons")) {
			if(type.endsWith("=vertical")) {
				panel.setLayout(new GridLayout(0, 1));
			}
			else {
				panel.setLayout(new GridLayout(1, 0));
			}

			ButtonGroup group = new ButtonGroup();

			for(int i = 0; i < coBoElements.size(); i += 1) {
				final String enumElem = coBoElements.get(i);

				JRadioButton radioButton = new JRadioButton(enumElem);
				radioButton.addActionListener(new ActionListener() {
					public void actionPerformed(ActionEvent e) {
						xPrefGUI.valueChanged(pageName, nodeID, enumElem, index);
					}
				});

				if(i == selectIndex) {
					radioButton.setSelected(true);
				}

				group.add(radioButton);
				panel.add(radioButton);
			}
		}
		else { // combo box...
			final JComboBox optionElement = new JComboBox(coBoElements.toArray());
			optionElement.setSelectedIndex(selectIndex);
			optionElement.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent arg0) {
					String saveValue = (String) optionElement.getSelectedItem(); // get new state

					xPrefGUI.valueChanged(pageName, nodeID, saveValue, index);
				}
			});

			panel.add(optionElement);
		}

		return panel;
	}

	/**
	 * Internal method to get the Enum associated with the option node.
	 * 
	 * @param optionNode
	 *            the option node to get the Enum for
	 * 
	 * @return the Enum of the option node
	 * 
	 * @throws ClassNotFoundException
	 *             if the source is invalid
	 */
	private static LinkedList<String> getEnum(Node optionNode) throws ClassNotFoundException {
		// get source...
		String source = XMLUtils.getAttributeValue(optionNode, "source");

		// create empty list for JComboBox...
		LinkedList<String> coBoElements = new LinkedList<String>();

		if(source != null) {
			Class<?> c = Class.forName(source); // get class

			Enum<?>[] enumElements = (Enum[]) c.getEnumConstants();

			for(int i = 0; i < enumElements.length; ++i) {
				coBoElements.add(enumElements[i].toString());
			}
		}

		return coBoElements;
	}

	/**
	 * Static method to get the first value of the given id as instance of the
	 * proper data type.
	 * 
	 * @param id
	 *            the id to get the value for
	 * 
	 * @return the element as instance of the right data type
	 * 
	 * @throws XPathExpressionException
	 *             if there is a problem in a XPath expression
	 * @throws ClassNotFoundException
	 *             if the class for the data type is not found
	 * @throws IOException
	 * @throws ParserConfigurationException
	 * @throws SAXException
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 * @throws IllegalClassFormatException
	 * @throws NoSuchMethodException
	 * @throws NoSuchFieldException
	 * @throws IllegalArgumentException
	 * @throws SecurityException
	 */
	public static <T extends Enum<?>> T getFirstValue(String id) throws XPathExpressionException, ClassNotFoundException, SAXException, ParserConfigurationException, IOException, SecurityException, IllegalArgumentException, NoSuchFieldException, NoSuchMethodException, IllegalClassFormatException, IllegalAccessException, InvocationTargetException {
		LinkedList<T> list = EnumDatatype.getValues(id);

		return list.get(0);
	}

	/**
	 * Static method to get the values to the given id as instances of the
	 * proper data type.
	 * 
	 * @param id
	 *            the id to get the values for
	 * 
	 * @return the elements as a List of instances of the right data type
	 * 
	 * @throws XPathExpressionException
	 *             if there is a problem in a XPath expression
	 * @throws ClassNotFoundException
	 *             if the class for the data type is not found
	 * @throws IOException
	 * @throws ParserConfigurationException
	 * @throws SAXException
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 * @throws IllegalClassFormatException
	 * @throws NoSuchMethodException
	 * @throws NoSuchFieldException
	 * @throws IllegalArgumentException
	 * @throws SecurityException
	 */
	@SuppressWarnings({ "unchecked", "rawtypes" })
	public static <T extends Enum<?>> LinkedList<T> getValues(String id) throws XPathExpressionException, ClassNotFoundException, SAXException, ParserConfigurationException, IOException, SecurityException, IllegalArgumentException, NoSuchFieldException, NoSuchMethodException, IllegalClassFormatException, IllegalAccessException, InvocationTargetException {
		LinkedList<T> enumList = new LinkedList<T>();

		String source = "";

		String[] idparts = id.split("\\.");

		if(idparts.length == 2) {
			String sequenceName = SimpleXPath.getInstance().getString("/preferences/page//option[@id='" + idparts[0] + "']/@datatype", XPref.getInstance().getMenuXML());

			source = SimpleXPath.getInstance().getString("/preferences/sequence[@name='" + sequenceName + "']/option[@id='" + idparts[1] + "']/@source", XPref.getInstance().getMenuXML());
		}
		else {
			source = SimpleXPath.getInstance().getString("//option[@id='" + id + "']/@source", XPref.getInstance().getMenuXML());
		}

		Class<T> c = (Class<T>) Class.forName(source); // get Enum class

		for(Object obj : XPref.getInstance().getValues(id)) {
			enumList.add((T) Enum.valueOf((Class) c, (String) obj));
		}

		return enumList;
	}

	/**
	 * Static method to describe the data type.
	 * 
	 * @return Description of the data type.
	 */
	public static String description() {
		return "Element out of a list of Enum values";
	}

	/**
	 * Static method to describe the usage for this option.
	 * 
	 * @param optionNode
	 *            the option node to show the description for
	 * 
	 * @return the description string
	 * 
	 * @throws ClassNotFoundException
	 *             if the class for the data type is not found
	 */
	public static String usage(Node optionNode) throws ClassNotFoundException {
		StringBuffer specific = new StringBuffer();

		// get source...
		String source = XMLUtils.getAttributeValue(optionNode, "source");

		if(source != null) {
			Class<?> c = Class.forName(source); // get class

			Enum<?>[] enumElements = (Enum[]) c.getEnumConstants();

			for(int i = 0; i < enumElements.length; i += 1) {
				specific.append(enumElements[i].toString());

				if(i < enumElements.length-1) {
					specific.append(" -- ");
				}
			}
		}

		return specific.toString();
	}

	/**
	 * Static method to get the name for this data type. The name can depend on
	 * some properties of the node it is used for.
	 * 
	 * @param optionNode
	 *            the node this data type is used for
	 * 
	 * @return the name of the data type
	 */
	public static String getName(Node optionNode) {
		// get source...
		String source = XMLUtils.getAttributeValue(optionNode, "source");

		return EnumDatatype.class.getName() + " (" + source + ")";
	}
}