package xpref.datatypes;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.IOException;
import java.lang.instrument.IllegalClassFormatException;
import java.lang.reflect.InvocationTargetException;
import java.util.LinkedList;

import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.xpath.XPathExpressionException;

import org.w3c.dom.Node;
import org.xml.sax.SAXException;

import xpref.XPref;
import xpref.XPrefGUI;
import xpref.util.XMLUtils;

public class FileChooseDatatype {
	/**
	 * The default value for this data type.
	 */
	protected static String defaultValue = "";

	/**
	 * Static method to serialize the given element.
	 * 
	 * @param element
	 *            the element to serialize
	 * 
	 * @return the serialized element as string
	 */
	public static String serializeElement(Object element) {
		if(element == null) {
			element = FileChooseDatatype.defaultValue;
		}

		return (String) element;
	}

	/**
	 * Static method to parse the string representation of the element to the
	 * right data type.
	 * 
	 * @param optionNode
	 *            the option node to parse the element for
	 * @param element
	 *            the string representation of the element
	 * 
	 * @return the element as instance of the right data type
	 */
	public static String parseElement(Node optionNode, String element) {
		if(element == null) {
			return StringDatatype.defaultValue;
		}
		else {
			return element;
		}
	}

	/**
	 * Static method to get the visual representation of the element for
	 * XPrefGUI.
	 * 
	 * @param element
	 *            the element to represent
	 * @param optionNode
	 *            node of the element
	 * @param xPrefGUI
	 *            instance of the XPrefGUI
	 * @param pageName
	 *            name of the page this value is on
	 * 
	 * @return the panel holding the visual representation
	 * 
	 * @throws NoSuchFieldException
	 *             if the mode is invalid
	 * @throws IllegalAccessException
	 * @throws SecurityException
	 * @throws IllegalArgumentException
	 */
	public static JPanel getVisualRepresentation(final String nodeID, Object element, final Node optionNode, final XPrefGUI xPrefGUI, final String pageName, final Integer index) throws IllegalArgumentException, SecurityException, IllegalAccessException, NoSuchFieldException {
		final String elem = (element == null) ? FileChooseDatatype.defaultValue : (String) element;

		final JTextField optionLabel = new JTextField(elem, 30);
		optionLabel.setEditable(false);

		final JFileChooser chooser = new JFileChooser(elem);
		chooser.setFileSelectionMode(JFileChooser.FILES_AND_DIRECTORIES);

		String selectionMode = XMLUtils.getAttributeValue(optionNode, "mode");

		if(selectionMode == null || selectionMode.equals("")) {
			selectionMode = "FILES_AND_DIRECTORIES";
		}

		Class<JFileChooser> jfilechooserClass = JFileChooser.class;

		chooser.setFileSelectionMode(jfilechooserClass.getField(selectionMode).getInt(jfilechooserClass));

		JButton optionElement = new JButton("Select");
		optionElement.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				int ret = chooser.showDialog(xPrefGUI, "Select");

				if(ret == JFileChooser.APPROVE_OPTION) {
					String fileName = chooser.getSelectedFile().getAbsolutePath();

					xPrefGUI.valueChanged(pageName, nodeID, FileChooseDatatype.serializeElement(fileName), index);

					optionLabel.setText(fileName);
				}
			}
		});

		// wrap check box in a panel...
		JPanel panel = new JPanel();
		panel.add(optionLabel);
		panel.add(optionElement);

		return panel;
	}

	/**
	 * Static method to get the first value of the given id as instance of the
	 * proper data type.
	 * 
	 * @param id
	 *            the id to get the value for
	 * 
	 * @return the element as instance of the right data type
	 * 
	 * @throws IOException
	 * @throws ParserConfigurationException
	 * @throws SAXException
	 * @throws XPathExpressionException
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 * @throws IllegalClassFormatException
	 * @throws NoSuchMethodException
	 * @throws NoSuchFieldException
	 * @throws ClassNotFoundException
	 * @throws IllegalArgumentException
	 * @throws SecurityException
	 */
	public static String getFirstValue(String id) throws XPathExpressionException, SAXException, ParserConfigurationException, IOException, SecurityException, IllegalArgumentException, ClassNotFoundException, NoSuchFieldException, NoSuchMethodException, IllegalClassFormatException, IllegalAccessException, InvocationTargetException {
		LinkedList<String> list = FileChooseDatatype.getValues(id);

		return list.get(0);
	}

	/**
	 * Static method to get the values to the given id as instances of the
	 * proper data type.
	 * 
	 * @param id
	 *            the id to get the values for
	 * 
	 * @return the elements as list of instances of the right data type
	 * 
	 * @throws IOException
	 * @throws ParserConfigurationException
	 * @throws SAXException
	 * @throws XPathExpressionException
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 * @throws IllegalClassFormatException
	 * @throws NoSuchMethodException
	 * @throws NoSuchFieldException
	 * @throws ClassNotFoundException
	 * @throws IllegalArgumentException
	 * @throws SecurityException
	 */
	public static LinkedList<String> getValues(String id) throws XPathExpressionException, SAXException, ParserConfigurationException, IOException, SecurityException, IllegalArgumentException, ClassNotFoundException, NoSuchFieldException, NoSuchMethodException, IllegalClassFormatException, IllegalAccessException, InvocationTargetException {
		LinkedList<String> list = new LinkedList<String>();

		for(Object obj : XPref.getInstance().getValues(id)) {
			list.add((String) obj);
		}

		return list;
	}

	/**
	 * Static method to describe the data type.
	 * 
	 * @return Description of the data type.
	 */
	public static String description() {
		return "Path to a file or directory";
	}

	/**
	 * Static method to describe the usage for this option.
	 * 
	 * @param optionNode
	 *            the option node to show the description for
	 * 
	 * @return the description string
	 */
	public static String usage(Node optionNode) {
		return "any string value";
	}

	/**
	 * Static method to get the name for this data type. The name can depend on
	 * some properties of the node it is used for.
	 * 
	 * @param optionNode
	 *            the node this data type is used for
	 * 
	 * @return the name of the data type
	 */
	public static String getName(Node optionNode) {
		return FileChooseDatatype.class.getName();
	}
}