package xpref.datatypes;

import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.IOException;
import java.lang.instrument.IllegalClassFormatException;
import java.lang.reflect.InvocationTargetException;
import java.util.LinkedList;

import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.xpath.XPathExpressionException;

import org.w3c.dom.Node;
import org.xml.sax.SAXException;

import xpref.XPref;
import xpref.XPrefGUI;
import xpref.util.JFontChooser;

public class FontDatatype {
	/**
	 * The default value for this data type.
	 */
	protected static Font defaultValue = new Font("Courier New", Font.PLAIN, 12);

	/**
	 * Static method to serialize the given element.
	 * 
	 * @param element
	 *            the element to serialize
	 * 
	 * @return the serialized element as string
	 */
	public static String serializeElement(Object element) {
		if(element == null) {
			element = FontDatatype.defaultValue;
		}

		Font elem = (Font) element;

		StringBuffer saveValue = new StringBuffer(elem.getFontName() + "; " + elem.getSize());

		if(elem.isItalic() || elem.isBold()) {
			saveValue.append("; ");

			if(elem.isItalic()) {
				saveValue.append(" italic");
			}

			if(elem.isBold()) {
				saveValue.append(" bold");
			}
		}

		return saveValue.toString();
	}

	/**
	 * Static method to parse the string representation of the element to the
	 * right data type.
	 * 
	 * @param optionNode
	 *            the option node to parse the element for
	 * @param element
	 *            the string representation of the element
	 * 
	 * @return the element as instance of the right data type
	 */
	public static Font parseElement(Node optionNode, String element) {
		if(element == null) {
			return FontDatatype.defaultValue;
		}
		else {
			String[] fontParts = element.split("; ");

			String fontName = fontParts[0];
			int fontSize = Integer.parseInt(fontParts[1]);
			int fontStyle = Font.PLAIN;

			if(fontParts.length == 3) {
				if(fontParts[2].toLowerCase().contains("bold")) {
					fontStyle |= Font.BOLD;
				}

				if(fontParts[2].toLowerCase().contains("italic")) {
					fontStyle |= Font.ITALIC;
				}
			}

			return new Font(fontName, fontStyle, fontSize);
		}
	}

	/**
	 * Static method to get the visual representation of the element for
	 * XPrefGUI.
	 * 
	 * @param element
	 *            the element to represent
	 * @param optionNode
	 *            node of the element
	 * @param xPrefGUI
	 *            instance of the XPrefGUI
	 * @param pageName
	 *            name of the page this value is on
	 * 
	 * @return the panel holding the visual representation
	 */
	public static JPanel getVisualRepresentation(final String nodeID, final Object element, Node optionNode, final XPrefGUI xPrefGUI, final String pageName, final Integer index) {
		final JLabel fontLabel = new JLabel(FontDatatype.serializeElement(element));

		JPanel optionElement = new JPanel();
		optionElement.add(fontLabel);

		JButton button = new JButton("choose");
		button.addActionListener(new ActionListener() {
			private Font font = (Font) element;

			public void actionPerformed(ActionEvent e) {
				Font newFont = JFontChooser.showDialog(xPrefGUI, "Choose Font", "Sample Text", this.font);

				if(newFont != null) {
					String saveValue = FontDatatype.serializeElement(newFont);

					fontLabel.setText(saveValue);

					this.font = newFont;

					xPrefGUI.valueChanged(pageName, nodeID, saveValue, index);
				}
			}
		});

		optionElement.add(button);

		// wrap check box in a panel...
		JPanel panel = new JPanel();
		panel.add(optionElement);

		return panel;
	}

	/**
	 * Static method to get the first value of the given id as instance of the
	 * proper data type.
	 * 
	 * @param id
	 *            the id to get the value for
	 * 
	 * @return the element as instance of the right data type
	 * 
	 * @throws IOException
	 * @throws ParserConfigurationException
	 * @throws SAXException
	 * @throws XPathExpressionException
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 * @throws IllegalClassFormatException
	 * @throws NoSuchMethodException
	 * @throws NoSuchFieldException
	 * @throws ClassNotFoundException
	 * @throws IllegalArgumentException
	 * @throws SecurityException
	 */
	public static Font getFirstValue(String id) throws XPathExpressionException, SAXException, ParserConfigurationException, IOException, SecurityException, IllegalArgumentException, ClassNotFoundException, NoSuchFieldException, NoSuchMethodException, IllegalClassFormatException, IllegalAccessException, InvocationTargetException {
		LinkedList<Font> list = FontDatatype.getValues(id);

		return list.get(0);
	}

	/**
	 * Static method to get the values to the given id as instance of the proper
	 * data type.
	 * 
	 * @param id
	 *            the id to get the values for
	 * 
	 * @return the elements as list of instances of the right data type
	 * 
	 * @throws IOException
	 * @throws ParserConfigurationException
	 * @throws SAXException
	 * @throws XPathExpressionException
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 * @throws IllegalClassFormatException
	 * @throws NoSuchMethodException
	 * @throws NoSuchFieldException
	 * @throws ClassNotFoundException
	 * @throws IllegalArgumentException
	 * @throws SecurityException
	 */
	public static LinkedList<Font> getValues(String id) throws XPathExpressionException, SAXException, ParserConfigurationException, IOException, SecurityException, IllegalArgumentException, ClassNotFoundException, NoSuchFieldException, NoSuchMethodException, IllegalClassFormatException, IllegalAccessException, InvocationTargetException {
		LinkedList<Font> fontList = new LinkedList<Font>();

		for(Object obj : XPref.getInstance().getValues(id)) {
			fontList.add((Font) obj);
		}

		return fontList;
	}

	/**
	 * Static method to describe the data type.
	 * 
	 * @return Description of the data type.
	 */
	public static String description() {
		return "Font representation";
	}

	/**
	 * Static method to describe the usage for this option.
	 * 
	 * @param optionNode
	 *            the option node to show the description for
	 * 
	 * @return the description string
	 */
	public static String usage(Node optionNode) {
		return "e.g.: \"Courier New; 12\"";
	}

	/**
	 * Static method to get the name for this data type. The name can depend on
	 * some properties of the node it is used for.
	 * 
	 * @param optionNode
	 *            the node this data type is used for
	 * 
	 * @return the name of the data type
	 */
	public static String getName(Node optionNode) {
		return BooleanDatatype.class.getName();
	}
}