package xpref.datatypes;

import java.awt.Dimension;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.io.IOException;
import java.lang.instrument.IllegalClassFormatException;
import java.lang.reflect.InvocationTargetException;
import java.util.LinkedList;

import javax.swing.JPanel;
import javax.swing.JSpinner;
import javax.swing.SpinnerNumberModel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.xpath.XPathExpressionException;

import org.w3c.dom.Node;
import org.xml.sax.SAXException;

import xpref.XPref;
import xpref.XPrefGUI;
import xpref.util.XMLUtils;

public class IntegerDatatype {
	/**
	 * The default value for this data type.
	 */
	protected static Integer defaultValue = new Integer(1);

	/**
	 * Static method to serialize the given element.
	 * 
	 * @param element
	 *            the element to serialize
	 * 
	 * @return the serialized element as string
	 */
	public static String serializeElement(Object element) {
		if(element == null) {
			element = IntegerDatatype.defaultValue;
		}

		return ((Integer) element).toString();
	}

	/**
	 * Static method to parse the string representation of the element to the
	 * right data type.
	 * 
	 * @param optionNode
	 *            the option node to parse the element for
	 * @param element
	 *            the string representation of the element
	 * 
	 * @return the element as instance of the right data type
	 */
	public static Integer parseElement(Node optionNode, String element) {
		if(element == null) {
			return IntegerDatatype.defaultValue;
		}
		else {
			return new Integer(element);
		}
	}

	/**
	 * Static method to get the visual representation of the element for
	 * XPrefGUI.
	 * 
	 * @param element
	 *            the element to represent
	 * @param optionNode
	 *            node of the element
	 * @param xPrefGUI
	 *            instance of the XPrefGUI
	 * @param pageName
	 *            name of the page this value is on
	 * 
	 * @return the panel holding the visual representation
	 */
	public static JPanel getVisualRepresentation(final String nodeID, Object element, Node optionNode, final XPrefGUI xPrefGUI, final String pageName, final Integer index) {
		// get minimum value...
		String minStr = XMLUtils.getAttributeValue(optionNode, "minimum");
		Integer minValue = null;

		if(minStr != null) {
			minValue = Integer.parseInt(minStr);
		}

		// get maximum value...
		String maxStr = XMLUtils.getAttributeValue(optionNode, "maximum");
		Integer maxValue = null;

		if(maxStr != null) {
			maxValue = Integer.parseInt(maxStr);
		}

		Integer elem = null;

		if(element == null) {
			elem = IntegerDatatype.defaultValue;
		}
		else {
			elem = (Integer) element;
		}

		// create spinner for the integer with the minimum and maximum values...
		final JSpinner optionElement = new JSpinner(new SpinnerNumberModel(elem, minValue, maxValue, new Integer(1)));
		optionElement.addChangeListener(new ChangeListener() {
			public void stateChanged(ChangeEvent fe) {
				IntegerDatatype.processStateChange(optionElement, xPrefGUI, pageName, nodeID, index);
			}
		});
		optionElement.addFocusListener(new FocusListener() {
			public void focusGained(FocusEvent fe) {}

			public void focusLost(FocusEvent fe) {
				IntegerDatatype.processStateChange(optionElement, xPrefGUI, pageName, nodeID, index);
			}
		});
		optionElement.setPreferredSize(new Dimension(optionElement.getPreferredSize().width + 10, optionElement.getPreferredSize().height));

		// wrap spinner in a panel...
		JPanel panel = new JPanel();
		panel.add(optionElement);

		return panel;
	}

	private static void processStateChange(JSpinner optionElement, XPrefGUI xPrefGUI, String pageName, String nodeID, Integer index) {
		String saveValue = ((Integer) optionElement.getValue()).toString();

		xPrefGUI.valueChanged(pageName, nodeID, saveValue, index);
	}

	/**
	 * Static method to get the first value of the given id as instance of the
	 * proper data type.
	 * 
	 * @param id
	 *            the id to get the value for
	 * 
	 * @return the element as instance of the right data type
	 * 
	 * @throws IOException
	 * @throws ParserConfigurationException
	 * @throws SAXException
	 * @throws XPathExpressionException
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 * @throws IllegalClassFormatException
	 * @throws NoSuchMethodException
	 * @throws NoSuchFieldException
	 * @throws ClassNotFoundException
	 * @throws IllegalArgumentException
	 * @throws SecurityException
	 */
	public static Integer getFirstValue(String id) throws XPathExpressionException, SAXException, ParserConfigurationException, IOException, SecurityException, IllegalArgumentException, ClassNotFoundException, NoSuchFieldException, NoSuchMethodException, IllegalClassFormatException, IllegalAccessException, InvocationTargetException {
		LinkedList<Integer> list = IntegerDatatype.getValues(id);

		return list.get(0);
	}

	/**
	 * Static method to get the values to the given id as instance of the proper
	 * data type.
	 * 
	 * @param id
	 *            the id to get the values for
	 * 
	 * @return the elements as a list of instances of the right data type
	 * 
	 * @throws IOException
	 * @throws ParserConfigurationException
	 * @throws SAXException
	 * @throws XPathExpressionException
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 * @throws IllegalClassFormatException
	 * @throws NoSuchMethodException
	 * @throws NoSuchFieldException
	 * @throws ClassNotFoundException
	 * @throws IllegalArgumentException
	 * @throws SecurityException
	 */
	public static LinkedList<Integer> getValues(String id) throws XPathExpressionException, SAXException, ParserConfigurationException, IOException, SecurityException, IllegalArgumentException, ClassNotFoundException, NoSuchFieldException, NoSuchMethodException, IllegalClassFormatException, IllegalAccessException, InvocationTargetException {
		LinkedList<Integer> integerList = new LinkedList<Integer>();

		for(Object obj : XPref.getInstance().getValues(id)) {
			integerList.add((Integer) obj);
		}

		return integerList;
	}

	/**
	 * Static method to describe the data type.
	 * 
	 * @return Description of the data type.
	 */
	public static String description() {
		return "Integer";
	}

	/**
	 * Static method to describe the usage for this option.
	 * 
	 * @param optionNode
	 *            the option node to show the description for
	 * 
	 * @return the description string
	 */
	public static String usage(Node optionNode) {
		StringBuffer specific = new StringBuffer();

		String minValue = XMLUtils.getAttributeValue(optionNode, "minimum");
		String maxValue = XMLUtils.getAttributeValue(optionNode, "maximum");

		if(minValue != null) {
			specific.append("Minimum: " + minValue);
		}

		if(maxValue != null) {
			if(minValue != null) {
				specific.append(";");
			}

			specific.append(" Maximum: " + maxValue);
		}

		return specific.toString();
	}

	/**
	 * Static method to get the name for this data type. The name can depend on
	 * some properties of the node it is used for.
	 * 
	 * @param optionNode
	 *            the node this data type is used for
	 * 
	 * @return the name of the data type
	 */
	public static String getName(Node optionNode) {
		StringBuffer specific = new StringBuffer(IntegerDatatype.class.getName());

		String minValue = XMLUtils.getAttributeValue(optionNode, "minimum");
		String maxValue = XMLUtils.getAttributeValue(optionNode, "maximum");

		if(minValue != null || maxValue != null) {
			specific.append(" (");

			if(minValue != null) {
				specific.append("min: " + minValue);
			}

			if(maxValue != null) {
				if(minValue != null) {
					specific.append(";");
				}

				specific.append(" max: " + maxValue);
			}

			specific.append(")");
		}

		return specific.toString();
	}
}