package xpref.formatter;

import java.lang.reflect.InvocationTargetException;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.regex.Pattern;

import javax.xml.xpath.XPathExpressionException;

import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import xpref.XPref;
import xpref.util.SimpleXPath;
import xpref.util.Tupel;
import xpref.util.XMLUtils;

public class HTMLFormatter implements IFormatter {
	private XPref xPref = null;
	private StringBuffer content;
	private boolean shortenDataTypes = false;
	private String title = "XPref Helpall";
	private String cssContent = "";
	private HashMap<String, String> dataTypes = new HashMap<String, String>();
	private Map<String, String> blocks = new HashMap<String, String>();

	/**
	 * Constructor which takes an instance of the XPref class as argument.
	 * 
	 * @param xPref
	 *            instance of the XPref class
	 */
	public HTMLFormatter(XPref xPref) {
		this.xPref = xPref;

		this.reset();
	}

	/**
	 * Constructor which takes an instance of the XPref class and the title of
	 * the HTML page as argument.
	 * 
	 * @param xPref
	 *            instance of the XPref class
	 * @param title
	 *            the title of the HTML page
	 */
	public HTMLFormatter(XPref xPref, String title) {
		this(xPref, title, "");
	}

	/**
	 * Constructor which takes an instance of the XPref class, the title of the
	 * HTML page and some additional CSS definitions as argument.
	 * 
	 * @param xPref
	 *            instance of the XPref class
	 * @param title
	 *            the title of the HTML page
	 * @param cssContent
	 *            additional CSS definitions
	 */
	public HTMLFormatter(XPref xPref, String title, String cssContent) {
		this.title = title;
		this.cssContent = cssContent;
		this.xPref = xPref;

		this.reset();
	}

	/**
	 * Determines whether the formatter will shorten the data type definitions
	 * (default = false).
	 * 
	 * @param newValue
	 *            the new state of this option
	 */
	public void setShortenDataTypes(boolean newValue) {
		this.shortenDataTypes = newValue;
	}

	/**
	 * Method to reset the formatter to make it reusable.
	 */
	public void reset() {
		this.content = new StringBuffer();

		this.content.append("<html>\n" +
				"<style type='text/css'>\n" +
				"<!--\n" +
				"div.idlabel {\n" +
				"  font-size: 18px;\n" +
				"  font-weight: bold;\n" +
				"  margin-bottom: 3px;\n" +
				"}\n" +
				"div.datatypeLabel {\n" +
				"  font-size: 18px;\n" +
				"  font-weight: bold;\n" +
				"  margin-bottom: 3px;\n" +
				"}\n" +
				"\n" +
				"th {\n" +
				"  background-color: #EEEEEE;\n" +
				"  white-space: nowrap;\n" +
				"  vertical-align: top;\n" +
				"}\n" +
				"td, th {\n" +
				"  padding: 0 5px;\n" +
				"  text-align: left;\n" +
				"}\n" +
				"table.blockDefinition {\n" +
				"  border: 1px solid #000000;\n" +
				"  margin-left: 10px;\n" +
				"}\n" +
				"table.blockDefinition th {\n" +
				"  border-top: 1px solid #FFFFFF;\n" +
				"  border-bottom: 1px solid #FFFFFF;\n" +
				"}\n" +
				"td.content {\n" +
				"  padding: 0;\n" +
				"}\n" +
				"td.content > table {\n" +
				"  width: 100%;\n" +
				"}\n" +
				"td.value {\n" +
				"  border-bottom: 1px solid #EEEEEE;\n" +
				"  width: 100%;\n" +
				"}\n" +
				"tr.block > td {\n" +
				"  border: 1px solid #000000;\n" +
				"}\n" +
				this.cssContent +
				"-->\n" +
				"</style>\n" +
				"<title>" + this.title + "</title>\n" +
				"<body>\n\n" +
				"<h1>" + this.title + "</h1>\n\n"
		);
	}

	/**
	 * Method to add text at the top of the output.
	 * 
	 * @param text
	 *            the text to add
	 */
	public void addText(String text) {
		this.parseAndAdd(text);
	}

	/**
	 * Method to add a new block for an option.
	 * 
	 * @param ID
	 *            the option id
	 * @param name
	 *            the option name
	 * @param dataType
	 *            the data type of the option
	 * @param defaultValue
	 *            the default value of the option
	 * @param helpText
	 *            the help text of the option
	 * @param possibleValues
	 *            the possible values of the option
	 * @param description
	 *            the description of the option
	 * 
	 * @throws XPathExpressionException
	 *             if there is an error in the XPath expressions
	 * @throws NoSuchMethodException
	 *             if the data type is invalid
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 * @throws ClassNotFoundException
	 * @throws SecurityException
	 * @throws IllegalArgumentException
	 */
	public void newBlock(String ID, String name, String dataType, String defaultValue, String helpText, String possibleValues, String description) throws XPathExpressionException, IllegalArgumentException, SecurityException, ClassNotFoundException, IllegalAccessException, InvocationTargetException, NoSuchMethodException {
		this.newBlock(ID, name, dataType, defaultValue, helpText, possibleValues, description, new LinkedList<Tupel<Class<?>, String>>());
	}

	/**
	 * Method to add a new block for an option.
	 * 
	 * @param ID
	 *            the option id
	 * @param name
	 *            the option name
	 * @param dataType
	 *            the data type of the option
	 * @param defaultValue
	 *            the default value of the option
	 * @param helpText
	 *            the help text of the option
	 * @param possibleValues
	 *            the possible values of the option
	 * @param description
	 *            the description of the option
	 * @param possibleSequenceValues
	 *            possible values for the option (only defined if the option is
	 *            a sequence
	 * 
	 * @throws XPathExpressionException
	 *             if there is an error in the XPath expressions
	 * @throws NoSuchMethodException
	 *             if the data type is invalid
	 * @throws IllegalArgumentException
	 * @throws SecurityException
	 * @throws ClassNotFoundException
	 * @throws IllegalAccessException
	 * @throws InvocationTargetException
	 */
	public void newBlock(String ID, String name, String dataType, String defaultValue, String helpText, String possibleValues, String description, LinkedList<Tupel<Class<?>, String>> possibleSequenceValues) throws XPathExpressionException, IllegalArgumentException, SecurityException, ClassNotFoundException, IllegalAccessException, InvocationTargetException, NoSuchMethodException {
		Node optionNode = null;
		String dataTypeName = null;

		if(!possibleValues.contains(this.xPref.getSequenceSeparator())) {
			optionNode = SimpleXPath.getInstance().getNode("/preferences//page//option[@id='" + ID + "']", this.xPref.getMenuXML());
			dataTypeName = this.getDataTypeName(dataType, optionNode);
		}
		else {
			optionNode = SimpleXPath.getInstance().getNode("/preferences/sequence[@name='" + dataType + "']", this.xPref.getMenuXML());
			dataTypeName = dataType;
		}

		// --- create id definition - begin ---
		this.blocks.put(ID, "<div class='idlabel'>" + ID + "</div>\n" +
				"<table class='blockDefinition' cellspacing='0' cellpadding='0'>\n" +
				"<tr><th>Short description:</th><td class='value'>" + name + "</td></tr>\n" +
				"<tr><th>Data type:</th><td class='value'><a href='#" + dataTypeName + "'>" + dataTypeName + "</a></td></tr>\n" +
				"<tr><th>Default value:</th><td class='value'>" + defaultValue + "</td></tr>\n" +
				"<tr><th>Help:</th><td>" + helpText + "</td></tr>\n" +
				"</table>\n"
		);
		// --- create id definition - end ---

		// --- create data type definition - begin ---
		this.parseDataTypeDefinition(dataType, description, possibleValues, optionNode);
		// --- create data type definition - begin ---
	}

	/**
	 * Method to get the formatted content.
	 * 
	 * @return the formatted content
	 */
	public String getContent() {
		// --- sort and add id definitions to content - begin ---
		SortedSet<String> sortedset = new TreeSet<String>(this.blocks.keySet());
		int i = 0;

		for(String id : sortedset) {
			this.content.append(this.blocks.get(id));

			if(i != sortedset.size()-1) {
				this.content.append("\n<br />\n\n");
			}

			i += 1;
		}
		// --- sort and add id definitions to content - end ---

		// --- sort and add data type definitions to content - begin ---
		this.content.append("\n<br />\n" +
				"<br />\n\n" +
				"<h2>Data types:</h2>\n"
		);

		sortedset = new TreeSet<String>(this.dataTypes.keySet());
		i = 0;

		for(String dataType : sortedset) {
			this.content.append(this.dataTypes.get(dataType));

			if(i != sortedset.size()-1) {
				this.content.append("\n<br />\n\n");
			}

			i += 1;
		}
		// --- sort and add data type definitions to content - end ---

		this.content.append("\n</body>\n" +
				"</html>\n"
		);

		return this.content.toString();
	}

	private void parseAndAdd(String text) {
		this.content.append(text.replace("\n", "<br />\n"));
	}

	private String shortenDataType(String dataType) {
		if(this.shortenDataTypes) {
			// delete all but the part after the last "." from the data type string...
			dataType = dataType.replaceAll("^(\\w|\\.)*\\.", "");
		}

		return dataType;
	}

	private String getSource(String dataType, Node datatypeNode) {
		String source = "";

		if(dataType.equals("xpref.datatypes.CollectionDatatype") || dataType.equals("xpref.datatypes.EnumDatatype")) {
			source = XMLUtils.getAttributeValue(datatypeNode, "source");

			if(!source.equals("")) {
				source = "<tr><th>Source:</th><td class='value'>" + source + "</td></tr>\n";
			}
		}

		return source;
	}

	/**
	 * Internal method to parse a data type definition.
	 * 
	 * @param dataType the data type
	 * @param description description of the data type
	 * @param possibleValues possible values of the data type
	 * @param optionNode option node for the current context
	 * 
	 * @throws ClassNotFoundException if the data type of the possible values is wrong
	 * @throws NoSuchMethodException if the data type is not valid
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 * @throws SecurityException
	 * @throws IllegalArgumentException
	 * @throws XPathExpressionException
	 */
	private void parseDataTypeDefinition(String dataType, String description, String possibleValues, Node optionNode) throws ClassNotFoundException, IllegalArgumentException, SecurityException, IllegalAccessException, InvocationTargetException, NoSuchMethodException, XPathExpressionException {
		String source = "";

		// -- deal with sequences - begin --
		String separator = this.xPref.getSequenceSeparator();

		if(possibleValues.contains(separator)) {
			description = XMLUtils.getAttributeValue(optionNode, "description");

			if(description == null) {
				description = "";
			}

			StringBuffer possibleValuesTmp = new StringBuffer();
			String[] parts = possibleValues.substring(1, possibleValues.length()-1).split(Pattern.quote(separator));

			// walk through the data types present in the sequence...
			for(int i = 0; i < parts.length; i += 1) {
				// get data type class...
				Class<?> partClass = Class.forName(parts[i]);

				// get description of the data type...
				String desc = (String) partClass.getMethod("description").invoke(partClass);

				// get option node...
				NodeList dataTypeNodes = SimpleXPath.getInstance().getNodeSet("//sequence[@name='" + dataType + "']/option", this.xPref.getMenuXML());
				Node datatypeNode = dataTypeNodes.item(i);

				String subDataType = XMLUtils.getAttributeValue(datatypeNode, "datatype");

				// get usage of the data type for the current node...
				String usage = (String) partClass.getMethod("usage", Node.class).invoke(partClass, datatypeNode);

				this.parseDataTypeDefinition(subDataType, desc, usage, datatypeNode);

				// get name of the data type for the current node...
				parts[i] = this.getDataTypeName(parts[i], datatypeNode);

				// add link around data type in the sequence...
				possibleValuesTmp.append("<a href='#" + parts[i] + "'>" + parts[i] + "</a>");

				if(i != parts.length-1) {
					possibleValuesTmp.append(separator);
				}
			}

			possibleValues = possibleValuesTmp.toString();
		}
		// -- deal with sequences - end --
		else {
			source = this.getSource(dataType, optionNode);
			dataType = getDataTypeName(dataType, optionNode);

		}

		this.dataTypes.put(dataType, "<div class='datatypeLabel'><a name='" + dataType + "'>" + dataType + "</a></div>\n" +
				"<table class='blockDefinition' cellspacing='0' cellpadding='0'>\n" +
				"<tr><th>Description:</th><td class='value'>" + description + "</td></tr>\n" +
				"<tr><th>Possible values:</th><td class='value'>" + possibleValues + "</td></tr>\n" +
				source +
				"</table>\n"
		);
	}

	/**
	 * Internal method to get the name of a data type.
	 * 
	 * @param dataType
	 *            the data type
	 * @param optionNode
	 *            the option node of the current context
	 * 
	 * @return the unique name of a data type
	 * 
	 * @throws ClassNotFoundException
	 *             if the dataType is not valid
	 * @throws NoSuchMethodException
	 *             if the dataType is not valid
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 * @throws SecurityException
	 * @throws IllegalArgumentException
	 */
	private String getDataTypeName(String dataType, Node optionNode) throws ClassNotFoundException, IllegalArgumentException, SecurityException, IllegalAccessException, InvocationTargetException, NoSuchMethodException {
		Class<?> clazz = Class.forName(dataType);

		return this.shortenDataType((String) clazz.getMethod("getName", Node.class).invoke(clazz, optionNode));
	}
}