package xpref.util;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Hashtable;
import java.util.LinkedList;

import javax.swing.JButton;
import javax.swing.JPanel;
import javax.swing.tree.TreePath;

public class History {
	/**
	 * Reference to itself.
	 */
	private History that = this;

	/**
	 * List of visited pages for the history.
	 */
	private LinkedList<String> list = new LinkedList<String>();

	/**
	 * Current page index in the history list.
	 */
	private int listIndex = -1;

	/**
	 * List of buttons to go one element forward in the page history for each page.
	 */
	private Hashtable<String, JButton> forwardButtons = new Hashtable<String, JButton>();

	/**
	 * List of buttons to go one element back in the page history for each page.
	 */
	private Hashtable<String, JButton> backButtons = new Hashtable<String, JButton>();

	/**
	 * The Tree instance of the GUI.
	 */
	private Tree tree = null;

	public History(Tree tree) {
		this.tree = tree;
	}

	/**
	 * Public method to clear the history.
	 */
	public void clear() {
		this.list.clear();
		this.listIndex = -1;

		for(JButton b : this.backButtons.values()) {
			b.setEnabled(false);
		}

		for(JButton b : this.forwardButtons.values()) {
			b.setEnabled(false);
		}
	}

	/**
	 * Public method to update the history.
	 * 
	 * @param pageID
	 *            the name of the current page
	 */
	public void update(String pageID) {
		// mark current page as unselected...
		if(this.listIndex != -1) {
			((TreeNode) this.tree.getTreePath(this.list.get(this.listIndex)).getLastPathComponent()).selected = false;
		}

		// check previous element..
		if(this.listIndex > 0 && this.list.get(this.listIndex-1).equals(pageID)) {
			this.listIndex--;
			this.forwardButtons.get(pageID).setEnabled(true);

			if(this.listIndex == 0) {
				this.backButtons.get(pageID).setEnabled(false);
			}
		}
		// check next element...
		else if(this.listIndex < this.list.size()-1 && this.list.get(this.listIndex+1).equals(pageID)) {
			this.listIndex++;
			this.backButtons.get(pageID).setEnabled(true);

			if(this.listIndex == this.list.size()-1) {
				this.forwardButtons.get(pageID).setEnabled(false);
			}
		}
		// new page...
		else {
			// delete all elements currently in the "future"...
			while(this.listIndex < this.list.size()-1) {
				this.list.removeLast();
			}

			// add new element
			this.list.add(pageID);
			this.listIndex++;

			if(this.listIndex > 0) {
				this.backButtons.get(pageID).setEnabled(true);
			}

			if(this.listIndex == this.list.size()-1) {
				this.forwardButtons.get(pageID).setEnabled(false);
			}
		}

		// select tree path of the new page
		TreePath treePath = this.tree.getTreePath(this.list.get(listIndex));
		((TreeNode) treePath.getLastPathComponent()).selected = true;
		this.tree.setSelectionPath(treePath);
	}

	/**
	 * Public method to update the history to the next entry.
	 */
	public void updateToNext() {
		this.update(this.list.get(this.listIndex+1));
	}

	/**
	 * Public method to update the history to the previous entry.
	 */
	public void updateToPrevious() {
		this.update(this.list.get(this.listIndex-1));
	}

	/**
	 * Public method to add back and forward button for a page.
	 * 
	 * @param pageID
	 *            the pageID to add the buttons for
	 * 
	 * @return the panel with the buttons
	 */
	public JPanel addButtons(String pageID) {
		this.backButtons.put(pageID, new JButton("<"));
		this.backButtons.get(pageID).setEnabled(false);
		this.backButtons.get(pageID).addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent arg0) {
				that.update(that.list.get(that.listIndex-1));
			}
		});

		this.forwardButtons.put(pageID, new JButton(">"));
		this.forwardButtons.get(pageID).setEnabled(false);
		this.forwardButtons.get(pageID).addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent arg0) {
				that.update(that.list.get(that.listIndex+1));
			}
		});

		JPanel historyPanel = new JPanel();
		historyPanel.add(this.backButtons.get(pageID));
		historyPanel.add(this.forwardButtons.get(pageID));

		return historyPanel;
	}
}