package xpref.util;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.Frame;
import java.awt.GraphicsEnvironment;
import java.awt.GridLayout;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;

/**
 * <p>
 * <b>JFontChooser</b> - aka OzFontChooser a font selection widget. Please note
 * this software is licensed under the LGPL. Please see <a
 * href="./LICENSE.txt">LICENSE.txt</a> distributed with this jar. This font
 * selection dialog is intended to work just like the java.awt.JColorChooser.
 * </p>
 * 
 * <p>
 * JFontChooser provides a pane of controls designed to allow a user to
 * manipulate and select a color.
 * </p>
 * 
 * <p>
 * This class provides three levels of API:<br>
 * 
 * 1. A static convenience method which shows a modal color-chooser dialog and
 * returns the color selected by the user.<br>
 * </p>
 * 
 * This class includes a main method for testing the widget. To run type<br>
 * <br>
 * <code>java -jar JFontChooser</code> <br>
 * <br>
 * to use in your app... put this jar in your classpath and then type<br>
 * <br>
 * <code>import com.ozten.font.JFontChooser;</code> <br>
 * <br>
 * then when you want to get a font from the user..<br>
 * <br>
 * <code>Font f = JFontChooser.showDialog(frame);</code> <br>
 * <br>
 * Please send suggestions, and bug reports to me
 * 
 * @author <a href="mailto:austinking@hotmail.com">Ozten</a> of <a
 *         href="http://www.cruftworks.com/software/">CruftWorks.com</a>
 * @version 0.21 OTHER TODOS and questions... <br>
 *          1 - Do I need to code up a JFontChooser.AccessibleJFontChooser inner
 *          class to implement accessibility support???<br>
 *          2 - implement commented out method public static JDialog
 *          createDialog...
 */
@SuppressWarnings("deprecation")
public class JFontChooser extends JComponent {
	private static final long serialVersionUID = 1L;
	private String sizes[] = { "8", "9", "10", "11", "12", "13", "14", "16", "18", "20", "24", "26", "28", "32", "36", "40", "48", "56", "64", "72" };
	private static String[] fonts;
	private JDialog jd;
	private JPanel p;
	private JComboBox fontChoices, sizeChoices;
	private JScrollPane previewScroll;
	private JTextArea preview;
	private static JButton apply, cancel;
	private Font fPreview;
	private JCheckBox checkBold, checkItalic;

	static {
		try {
			fonts = GraphicsEnvironment.getLocalGraphicsEnvironment().getAvailableFontFamilyNames();
		}
		catch(Throwable t) {
			fonts = Toolkit.getDefaultToolkit().getFontList();
		}
	}

	/**
	 * This is a convenience method for those whom don't want to pass a font, a
	 * sample String, and a Dialog title in.
	 * 
	 * @see #showDialog(Component, String, String, Font)
	 */
	public static Font showDialog(Component c) {
		return showDialog(c, "Choose a font", "How about this font?");
	}

	/**
	 * This is a convenience method for those whom don't want to pass a font, and
	 * a sample String in.
	 * 
	 * @see #showDialog(Component, String, String, Font)
	 */
	public static Font showDialog(Component c, String title) {
		return showDialog(c, title, "How about this font?", c.getFont());
	}

	/**
	 * This is a convenience method for those whom don't want to pass a font in.
	 * 
	 * @see #showDialog(Component, String, String, Font)
	 */
	public static Font showDialog(Component c, String title, String sampleText) {
		return showDialog(c, title, sampleText, c.getFont());
	}

	/**
	 * Shows a modal font chooser dialog. If user presses "OK" the dialog is
	 * disposed and the currently previewed font is returned. If the user
	 * selects "Cancel" then null is returned.
	 * 
	 * @param component
	 *            - the parent component for the Dialog.
	 * @param title
	 *            - the title of the window such as "Choose Font"
	 * @param sampleText
	 *            - the words to preview the font with
	 * @param initial
	 *            - the font to start previewing with. You should pass in the
	 *            current font being used for smooth user exp.
	 * @return Font you will get back either null if user cancels, or a new Font
	 *         that the user has choosen.
	 */
	public static Font showDialog(Component component, String title, String sampleText, Font initial) {
		final JFontChooser jfc = new JFontChooser(initial, sampleText);
		jfc.doJFCLayout();
		jfc.setJDialog(JFontChooser.initDialog(component, title, true));
		jfc.getJDialog().getContentPane().add(jfc, BorderLayout.CENTER);

		JPanel buttons = new JPanel(new GridLayout(0, 1));
		buttons.add(apply = new JButton("OK"));
		buttons.add(cancel = new JButton("Cancel"));

		apply.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				jfc.setPreviewFont(jfc.buildFont());

				if(jfc.getJDialog() != null) {
					jfc.getJDialog().dispose();
					jfc.getJDialog().setVisible(false);
				}
			}
		});

		cancel.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				jfc.setPreviewFont(null);

				if(jfc.getJDialog() != null) {
					jfc.getJDialog().dispose();
					jfc.getJDialog().setVisible(false);
				}
			}
		});

		JPanel buttonsOuter = new JPanel();
		buttonsOuter.setLayout(new BorderLayout());
		buttonsOuter.add(buttons, BorderLayout.NORTH);
		jfc.getJDialog().getContentPane().add(buttonsOuter, BorderLayout.EAST);
		jfc.getJDialog().pack();
		jfc.getJDialog().setResizable(false);
		jfc.getJDialog().setLocationRelativeTo(null);
		jfc.getJDialog().setVisible(true);

		return jfc.getPreviewFont();
	}

	private static JDialog initDialog(Component c, String title, boolean modal) {
		JDialog rv = null;

		Frame f = getFrame(c);

		if(f != null) {
			rv = new JDialog(f, title, modal);
		}
		else { // component wasn't in a Frame
			rv = new JDialog();
		}

		return rv;
	}

	/**
	 * Walks through a components heirarcy and tries to find top level<br>
	 * <code>java.awt.Frame</code> object.
	 * 
	 * @param c
	 *            - the component
	 * @return Frame - the frame it is living in, or null.
	 */
	private static Frame getFrame(Component c) {
		Frame rv = null;

		if(c instanceof Frame) {
			rv = (Frame) c;
		}

		Container test = null;
		Container parent = c.getParent();

		if(parent == null) { // Component wasn't in a GUI
			return rv;
		}

		while(!(parent instanceof Frame)) {
			test = parent.getParent();

			if(test == null) { // Component wasn't in a GUI
				return rv;
			}

			parent = test;
		}

		if(parent instanceof Frame) {
			rv = (Frame) parent;
		}

		return rv;
	}

	/**
	 * Create a font chooser with initial Font and text in preview window
	 * 
	 * @param init
	 *            - the initial Font to preview in chooser.
	 * @param sampleText
	 *            - the words to test the font on.
	 */
	public JFontChooser(Font init, String sampleText) {
		super();
		this.setPreviewFont(init);
		this.preview = new JTextArea(sampleText);
	}

	private void doJFCLayout() {
		this.setLayout(new BorderLayout());
		this.p = new JPanel(new FlowLayout());
		this.add(this.p, BorderLayout.CENTER);

		this.fontChoices = new JComboBox(fonts);

		// Set up initial Font or 0 if none...
		boolean initIsValid = false;

		if(this.getPreviewFont() != null) {
			String[] fontsTemp = fonts;

			for(int i = 0; i < fontsTemp.length; i++) {
				if(fontsTemp[i].equalsIgnoreCase(this.getPreviewFont().getFontName())) {
					this.fontChoices.setSelectedIndex(i);

					initIsValid = true;
				}
			}
		}

		if(!initIsValid) {
			this.fontChoices.setSelectedIndex(0);
		}

		this.fontChoices.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				updateGUI();
			}
		});

		this.p.add(this.fontChoices);

		this.sizeChoices = new JComboBox(this.sizes);

		initIsValid = false;

		// Set up initial Font size or the 12th size (28ish)
		if(this.getPreviewFont() != null) {
			for(int j = 0; j < this.sizes.length; j++) {
				if(Integer.parseInt(this.sizes[j]) == this.getPreviewFont().getSize()) {
					this.sizeChoices.setSelectedIndex(j);

					initIsValid = true;
				}
			}
		}

		if(!initIsValid) {
			this.sizeChoices.setSelectedIndex(12);
		}

		this.sizeChoices.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				updateGUI();
			}
		});

		this.p.add(this.sizeChoices);

		JPanel stylePanel = new JPanel(new GridLayout(1, 2));

		this.checkBold = new JCheckBox("Bold", this.getPreviewFont() != null && this.getPreviewFont().isBold());
		this.checkBold.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				updateGUI();
			}
		});

		stylePanel.add(this.checkBold);

		this.checkItalic = new JCheckBox("Italic", this.getPreviewFont() != null && this.getPreviewFont().isItalic());
		this.checkItalic.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				updateGUI();
			}
		});

		stylePanel.add(this.checkItalic);

		JPanel upperRight = new JPanel(new GridLayout());
		upperRight.add(stylePanel);

		this.p.add(upperRight);

		this.previewScroll = new JScrollPane(this.preview);
		this.previewScroll.setPreferredSize(new Dimension(400, 80));

		this.add(this.previewScroll, BorderLayout.SOUTH);
		this.updateGUI();
	}

	private Font buildFont() {
		Font rv = null;

		String s = (String) this.sizeChoices.getSelectedItem();

		if(s != null) {
			int i = Integer.parseInt(s);
			int style = Font.PLAIN;

			if(this.checkBold.isSelected() && this.checkItalic.isSelected()) {
				style = Font.BOLD;
				style |= Font.ITALIC;
			}
			else if(this.checkBold.isSelected()) {
				style = Font.BOLD;
			}
			else if(this.checkItalic.isSelected()) {
				style = Font.ITALIC;
			}

			return new Font((String) this.fontChoices.getSelectedItem(), style, i);
		}

		return rv;
	}

	private void updateGUI() {
		Font fnew = this.buildFont();

		if(fnew != null && this.preview != null) {
			this.preview.setFont(fnew);

			this.setPreviewFont(fnew);
		}
	}


	public JDialog getJDialog() {
		return this.jd;
	}

	public void setJDialog(JDialog d) {
		this.jd = d;
	}

	/**
	 * When using JFontChooser as a component, use this to get the currently
	 * configured font.
	 */
	public Font getPreviewFont() {
		return this.fPreview;
	}

	public void setPreviewFont(Font f) {
		this.fPreview = f;
	}
}