package xpref.util;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;

import javax.xml.XMLConstants;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;
import javax.xml.validation.Validator;

import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.xml.sax.SAXException;

public class XMLUtils {
	/**
	 * String that holds the XSD for the menu XML.
	 */
	private static String menuXSD = "<?xml version='1.0' encoding='utf-8' ?>"
		+ "<xs:schema xmlns:xs='http://www.w3.org/2001/XMLSchema'>"
		+ "  <xs:complexType name='pageType'>"
		+ "    <xs:sequence>"
		+ "      <xs:element ref='page' minOccurs='0' maxOccurs='unbounded' />"
		+ "      <xs:element name='options' type='optionsType' minOccurs='0' maxOccurs='1' />"
		+ "    </xs:sequence>"
		+ "    <xs:attribute name='id' type='xs:ID' use='required' />"
		+ "    <xs:attribute name='name' type='xs:string' use='required' />"
		+ "  </xs:complexType>"
		+ "  <xs:complexType name='optionsType'>"
		+ "    <xs:sequence minOccurs='0' maxOccurs='unbounded'>"
		+ "      <xs:choice>"
		+ "        <xs:element name='option' type='optionType' />"
		+ "        <xs:element name='optionbox'>"
		+ "          <xs:complexType>"
		+ "            <xs:complexContent>"
		+ "              <xs:extension base='optionsType'>"
		+ "                <xs:attribute name='name' type='xs:string' use='required' />"
		+ "                <xs:attribute name='dependsOn' type='xs:string' />"
		+ "              </xs:extension>"
		+ "            </xs:complexContent>"
		+ "          </xs:complexType>"
		+ "        </xs:element>"
		+ "        <xs:element name='line' />"
		+ "        <xs:element name='text' />"
		+ "      </xs:choice>"
		+ "    </xs:sequence>"
		+ "  </xs:complexType>"
		+ "  <xs:complexType name='optionType'>"
		+ "    <xs:attribute name='id' type='xs:ID' use='required' />"
		+ "    <xs:attribute name='dependsOn' type='xs:string' />"
		+ "    <xs:attribute name='name' type='xs:string' use='required' />"
		+ "    <xs:attribute name='datatype' type='xs:string' use='required' />"
		+ "    <xs:attribute name='value' type='xs:string' />"
		+ "    <xs:attribute name='source' type='xs:string' />"
		+ "    <xs:attribute name='mode'>"
		+ "      <xs:simpleType>"
		+ "        <xs:restriction base='xs:string'>"
		+ "          <xs:pattern value='FILES_ONLY|DIRECTORIES_ONLY|FILES_AND_DIRECTORIES' />"
		+ "        </xs:restriction>"
		+ "      </xs:simpleType>"
		+ "    </xs:attribute>"
		+ "    <xs:attribute name='minimum' type='xs:integer' />"
		+ "    <xs:attribute name='maximum' type='xs:integer' />"
		+ "    <xs:attribute name='help' type='xs:string' />"
		+ "    <xs:attribute name='visible'>"
		+ "      <xs:simpleType>"
		+ "        <xs:restriction base='xs:string'>"
		+ "          <xs:pattern value='gui|commandline' />"
		+ "        </xs:restriction>"
		+ "      </xs:simpleType>"
		+ "    </xs:attribute>"
		+ "    <xs:attribute name='parameters' type='xs:string' />"
		+ "  </xs:complexType>"
		+ "  <xs:element name='page' type='pageType' />"
		+ "  <xs:element name='preferences'>"
		+ "    <xs:complexType>"
		+ "      <xs:sequence>"
		+ "        <xs:element name='sequence' minOccurs='0' maxOccurs='unbounded'>"
		+ "          <xs:complexType>"
		+ "            <xs:sequence>"
		+ "              <xs:element name='option' type='optionType' maxOccurs='unbounded' />"
		+ "            </xs:sequence>"
		+ "            <xs:attribute name='name' type='xs:string' use='required' />"
		+ "            <xs:attribute name='description' type='xs:string' />"
		+ "          </xs:complexType>"
		+ "        </xs:element>"
		+ "        <xs:element name='page' type='pageType' minOccurs='1' maxOccurs='unbounded' />"
		+ "      </xs:sequence>"
		+ "    </xs:complexType>"
		+ "  </xs:element>"
		+ "</xs:schema>";
	private static String saveXSD = "<?xml version='1.0' encoding='UTF-8'?>"
		+ "<xs:schema xmlns:xs='http://www.w3.org/2001/XMLSchema'>"
		+ "  <xs:element name='savedPreferences'>"
		+ "    <xs:complexType>"
		+ "      <xs:sequence>"
		+ "        <xs:element name='item' minOccurs='0' maxOccurs='unbounded'>"
		+ "          <xs:complexType>"
		+ "            <xs:simpleContent>"
		+ "               <xs:extension base='xs:string'>"
		+ "                 <xs:attribute name='id' type='xs:ID' use='required' />"
		+ "               </xs:extension>"
		+ "            </xs:simpleContent>"
		+ "          </xs:complexType>"
		+ "        </xs:element>"
		+ "      </xs:sequence>"
		+ "    </xs:complexType>"
		+ "  </xs:element>"
		+ "</xs:schema>";

	/**
	 * Method to parse a menu XML which is specified by the given file name.
	 * 
	 * @param xmlFileName
	 *            file name of the menu XML
	 * 
	 * @return the parsed DOMDocument
	 * 
	 * @throws IOException
	 * @throws ParserConfigurationException
	 * @throws SAXException
	 */
	public static Document readMenuXML(String xmlFileName) throws SAXException, ParserConfigurationException, IOException {
		return XMLUtils.readXML(xmlFileName, XMLUtils.menuXSD);
	}

	/**
	 * Method to parse a menu XML which is specified by the given InputStream.
	 * 
	 * @param xml
	 *            InputStream of the menu XML
	 * 
	 * @return the parsed DOMDocument
	 * 
	 * @throws IOException
	 * @throws ParserConfigurationException
	 * @throws SAXException
	 */
	public static Document readMenuXML(InputStream xml) throws SAXException, ParserConfigurationException, IOException {
		return XMLUtils.readXML(xml, XMLUtils.menuXSD);
	}

	/**
	 * Method to parse a save XML which is specified by the given file name.
	 * 
	 * @param xmlFileName file name for the save XML
	 * 
	 * @return the parsed DOMDocument
	 * 
	 * @throws IOException
	 * @throws ParserConfigurationException
	 * @throws SAXException
	 */
	public static Document readSaveXML(final String xmlFileName) throws SAXException, ParserConfigurationException, IOException {
		return XMLUtils.readXML(xmlFileName, XMLUtils.saveXSD);
	}

	/**
	 * Internal method to parse a given XML file and validate it against a given
	 * XSD file.
	 * 
	 * @param xmlFileName
	 *            name of the XML file to parse and validate
	 * @param xsdFileName
	 *            name of the XSD file to validate against
	 * 
	 * @return the parsed DOMDocument
	 * 
	 * @throws IOException
	 * @throws ParserConfigurationException
	 * @throws SAXException
	 */
	private static Document readXML(String xmlFileName, String xsdFileName) throws SAXException, ParserConfigurationException, IOException {
		File xmlFile = new File(xmlFileName);

		if(!xmlFile.exists()) {
			throw new FileNotFoundException();
		}

		System.out.println("reading XML file: " + xmlFileName);

		return XMLUtils.readXML(new FileInputStream(xmlFile), xsdFileName);
	}

	/**
	 * Internal method to parse a XML given by an InputStream and validate it
	 * against a given XSD file.
	 * 
	 * @param xml
	 *            InputStream of the XML to parse and validate
	 * @param xsdFileName
	 *            name of the XSD file to validate against
	 * 
	 * @return the parsed DOMDocument
	 *
	 * @throws SAXException
	 * @throws ParserConfigurationException
	 * @throws IOException
	 */
	private static Document readXML(InputStream xml, String xsdFileName) throws SAXException, ParserConfigurationException, IOException {
		// build an XSD-aware SchemaFactory
		SchemaFactory schemaFactory = SchemaFactory.newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI);

		// get the custom XSD schema describing the required format for my
		// XML files.
		Schema schemaXSD = schemaFactory.newSchema(new StreamSource(new ByteArrayInputStream(xsdFileName.getBytes("UTF-8"))));

		// Create a validator capable of validating XML files according to
		// my custom schema.
		Validator validator = schemaXSD.newValidator();

		// Get a parser capable of parsing vanilla XML into a DOM tree
		DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
		dbf.setNamespaceAware(true);

		DocumentBuilder parser = dbf.newDocumentBuilder();

		// parse the XML purely as XML and get a DOM tree representation.
		Document xmlDoc = parser.parse(xml);

		// parse the XML DOM tree against the stricter XSD schema
		validator.validate(new DOMSource(xmlDoc));

		return xmlDoc;
	}

	/**
	 * Tries to retrieve a given attribute name for a node.
	 * 
	 * @param node
	 *            Node whose attributes are to be searched.
	 * @param attrName
	 *            Attribute whose value the caller is interested in.
	 * 
	 * @return Attribute value if existent, null otherwise
	 */
	public static String getAttributeValue(Node node, String attrName) {
		NamedNodeMap nnm = node.getAttributes();

		if(nnm == null) {
			return null;
		}

		Node n = nnm.getNamedItem(attrName);

		if(n == null) {
			return null;
		}

		return n.getNodeValue();
	}

	/**
	 * Method to write a XML file.
	 * 
	 * @param fileName
	 *            the file name
	 * @param xmlDocument
	 *            the DOMDocument
	 * 
	 * @throws IOException
	 * @throws TransformerException
	 */
	public static void writeXMLFile(String fileName, Document xmlDocument) throws IOException, TransformerException {
		File saveFile = new File(fileName);

		if(!saveFile.exists()) {
			saveFile.createNewFile();
		}

		Result result = new StreamResult(saveFile);
		Source source = new DOMSource(xmlDocument);

		TransformerFactory factory = TransformerFactory.newInstance();

		Transformer transformer = factory.newTransformer();
		transformer.setOutputProperty(OutputKeys.INDENT, "yes");
		transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "2");
		transformer.transform(source, result);
	}
}